"""Helpers for rendering span-era deployment summaries and trees."""

from collections import Counter
from dataclasses import dataclass
from datetime import timedelta
from typing import Any
from uuid import UUID

from ai_pipeline_core.database._json_helpers import parse_json_object
from ai_pipeline_core.database._types import (
    TOKENS_CACHE_READ_KEY,
    TOKENS_INPUT_KEY,
    TOKENS_OUTPUT_KEY,
    TOKENS_REASONING_KEY,
    CostTotals,
    SpanKind,
    SpanRecord,
    SpanStatus,
)
from ai_pipeline_core.database.filesystem._paths import run_directory_name, span_filename

__all__ = [
    "SpanTreeView",
    "build_span_tree_view",
    "format_span_overview_lines",
    "format_span_tree_lines",
    "generate_costs_from_tree",
    "generate_summary_from_tree",
]

_FOUR_SPACES = "    "
MAX_TREE_DEPTH = 100

MODEL_KEY = "model"
PURPOSE_KEY = "purpose"
CACHE_HIT_KEY = "cache_hit"
STEP_KEY = "step"
TOTAL_STEPS_KEY = "total_steps"
ROUND_INDEX_KEY = "round_index"
TOOL_NAME_KEY = "tool_name"
FLOW_PLAN_KEY = "flow_plan"

UNKNOWN_MODEL_LABEL = "(unknown)"

_SPAN_KIND_PRIORITY: dict[str, int] = {
    SpanKind.DEPLOYMENT: 0,
    SpanKind.FLOW: 1,
    SpanKind.TASK: 2,
    SpanKind.OPERATION: 3,
    SpanKind.CONVERSATION: 4,
    SpanKind.LLM_ROUND: 5,
    SpanKind.TOOL_CALL: 6,
}


@dataclass(frozen=True, slots=True)
class _TokenTotals:
    """Aggregated token counts for descendant LLM rounds."""

    tokens_input: int = 0
    tokens_output: int = 0
    tokens_cache_read: int = 0
    tokens_reasoning: int = 0


@dataclass(frozen=True, slots=True)
class SpanTreeView:
    """Precomputed indexes and totals for one span deployment tree."""

    root_span: SpanRecord
    spans_by_id: dict[UUID, SpanRecord]
    children_map: dict[UUID | None, list[UUID]]
    meta_by_id: dict[UUID, dict[str, Any]]
    metrics_by_id: dict[UUID, dict[str, Any]]
    counts_by_kind: Counter[str]
    document_shas: set[str]
    descendant_costs: dict[UUID, float]
    descendant_tokens: dict[UUID, _TokenTotals]
    totals: CostTotals


def _detail_int(payload: dict[str, Any], key: str, *, context: str, field_name: str) -> int:
    value = payload.get(key, 0)
    if value is None:
        return 0
    if isinstance(value, bool):
        return int(value)
    try:
        return int(value)
    except (TypeError, ValueError) as exc:
        msg = f"{context} {field_name}[{key!r}] must be an integer value. Persist an integer before rendering summaries."
        raise ValueError(msg) from exc


def _detail_str(payload: dict[str, Any], key: str) -> str:
    value = payload.get(key, "")
    if isinstance(value, str):
        return value
    return ""


def _detail_bool(payload: dict[str, Any], key: str) -> bool:
    return payload.get(key) is True


def _child_sort_key(span_id: UUID, spans_by_id: dict[UUID, SpanRecord]) -> tuple[Any, int, int, str]:
    span = spans_by_id[span_id]
    return (
        span.started_at,
        span.sequence_no,
        _SPAN_KIND_PRIORITY.get(span.kind, len(_SPAN_KIND_PRIORITY)),
        str(span.span_id),
    )


def _build_children_map(tree: list[SpanRecord]) -> dict[UUID | None, list[UUID]]:
    children_map: dict[UUID | None, list[UUID]] = {}
    spans_by_id = {span.span_id: span for span in tree}
    for span in tree:
        children_map.setdefault(span.parent_span_id, []).append(span.span_id)
    for parent_id, child_ids in children_map.items():
        children_map[parent_id] = sorted(child_ids, key=lambda child_id: _child_sort_key(child_id, spans_by_id))
    return children_map


def _select_root_span(tree: list[SpanRecord], root_deployment_id: UUID) -> SpanRecord | None:
    deployment_matches = [span for span in tree if span.kind == SpanKind.DEPLOYMENT and span.deployment_id == root_deployment_id]
    if deployment_matches:
        return min(deployment_matches, key=lambda span: (span.started_at, span.sequence_no, str(span.span_id)))

    top_level_matches = [span for span in tree if span.kind == SpanKind.DEPLOYMENT and span.parent_span_id is None]
    if top_level_matches:
        return min(top_level_matches, key=lambda span: (span.started_at, span.sequence_no, str(span.span_id)))

    deployment_spans = [span for span in tree if span.kind == SpanKind.DEPLOYMENT]
    if deployment_spans:
        return min(deployment_spans, key=lambda span: (span.started_at, span.sequence_no, str(span.span_id)))
    return None


def _compute_descendant_costs(
    *,
    spans_by_id: dict[UUID, SpanRecord],
    children_map: dict[UUID | None, list[UUID]],
    metrics_by_id: dict[UUID, dict[str, Any]],
) -> tuple[dict[UUID, float], dict[UUID, _TokenTotals]]:
    descendant_costs: dict[UUID, float] = {}
    descendant_tokens: dict[UUID, _TokenTotals] = {}
    visiting: set[UUID] = set()

    def _visit(span_id: UUID) -> tuple[float, _TokenTotals]:
        if span_id in descendant_costs:
            return descendant_costs[span_id], descendant_tokens[span_id]
        if span_id in visiting:
            return 0.0, _TokenTotals()

        visiting.add(span_id)
        total_cost = 0.0
        ti, to, tc, tr = 0, 0, 0, 0
        for child_id in children_map.get(span_id, []):
            child = spans_by_id[child_id]
            total_cost += child.cost_usd
            if child.kind == SpanKind.LLM_ROUND:
                metrics = metrics_by_id[child_id]
                context = f"Span {child_id}"
                ti += _detail_int(metrics, TOKENS_INPUT_KEY, context=context, field_name="metrics_json")
                to += _detail_int(metrics, TOKENS_OUTPUT_KEY, context=context, field_name="metrics_json")
                tc += _detail_int(metrics, TOKENS_CACHE_READ_KEY, context=context, field_name="metrics_json")
                tr += _detail_int(metrics, TOKENS_REASONING_KEY, context=context, field_name="metrics_json")
            child_cost, child_tokens = _visit(child_id)
            total_cost += child_cost
            ti += child_tokens.tokens_input
            to += child_tokens.tokens_output
            tc += child_tokens.tokens_cache_read
            tr += child_tokens.tokens_reasoning
        descendant_costs[span_id] = total_cost
        descendant_tokens[span_id] = _TokenTotals(tokens_input=ti, tokens_output=to, tokens_cache_read=tc, tokens_reasoning=tr)
        visiting.remove(span_id)
        return total_cost, descendant_tokens[span_id]

    for span_id in spans_by_id:
        _visit(span_id)
    return descendant_costs, descendant_tokens


def _sum_span_totals(tree: list[SpanRecord], metrics_by_id: dict[UUID, dict[str, Any]]) -> CostTotals:
    cost_usd = 0.0
    ti, to, tc, tr = 0, 0, 0, 0
    for span in tree:
        cost_usd += span.cost_usd
        if span.kind == SpanKind.LLM_ROUND:
            metrics = metrics_by_id[span.span_id]
            context = f"Span {span.span_id}"
            ti += _detail_int(metrics, TOKENS_INPUT_KEY, context=context, field_name="metrics_json")
            to += _detail_int(metrics, TOKENS_OUTPUT_KEY, context=context, field_name="metrics_json")
            tc += _detail_int(metrics, TOKENS_CACHE_READ_KEY, context=context, field_name="metrics_json")
            tr += _detail_int(metrics, TOKENS_REASONING_KEY, context=context, field_name="metrics_json")
    return CostTotals(cost_usd=cost_usd, tokens_input=ti, tokens_output=to, tokens_cache_read=tc, tokens_reasoning=tr)


def _collect_document_shas(tree: list[SpanRecord]) -> set[str]:
    shas: set[str] = set()
    for span in tree:
        shas.update(span.input_document_shas)
        shas.update(span.output_document_shas)
    return shas


def build_span_tree_view(tree: list[SpanRecord], root_deployment_id: UUID) -> SpanTreeView | None:
    """Build reusable indexes and totals for a span deployment tree."""
    if not tree:
        return None

    root_span = _select_root_span(tree, root_deployment_id)
    if root_span is None:
        return None

    spans_by_id = {span.span_id: span for span in tree}
    meta_by_id = {span.span_id: parse_json_object(span.meta_json, context=f"Span {span.span_id}", field_name="meta_json") for span in tree}
    metrics_by_id = {span.span_id: parse_json_object(span.metrics_json, context=f"Span {span.span_id}", field_name="metrics_json") for span in tree}
    children_map = _build_children_map(tree)
    descendant_costs, descendant_tokens = _compute_descendant_costs(
        spans_by_id=spans_by_id,
        children_map=children_map,
        metrics_by_id=metrics_by_id,
    )
    return SpanTreeView(
        root_span=root_span,
        spans_by_id=spans_by_id,
        children_map=children_map,
        meta_by_id=meta_by_id,
        metrics_by_id=metrics_by_id,
        counts_by_kind=Counter(span.kind for span in tree),
        document_shas=_collect_document_shas(tree),
        descendant_costs=descendant_costs,
        descendant_tokens=descendant_tokens,
        totals=_sum_span_totals(tree, metrics_by_id),
    )


def _format_timedelta(delta: timedelta) -> str:
    secs = delta.total_seconds()
    if secs < 1:
        return f"{int(secs * 1000)}ms"
    if secs < 60:
        return f"{secs:.1f}s" if secs < 10 else f"{secs:.0f}s"
    if secs < 3600:
        return f"{int(secs // 60)}m {int(secs % 60)}s"
    return f"{int(secs // 3600)}h {int((secs % 3600) // 60)}m"


def _format_duration(span: SpanRecord) -> str:
    if span.ended_at is None:
        return "running..." if span.status == SpanStatus.RUNNING else "-"
    return _format_timedelta(span.ended_at - span.started_at)


def _format_token_count(count: int) -> str:
    if count >= 1000:
        return f"{round(count / 1000)}K"
    return str(count)


def _format_token_parts(span: SpanRecord, metrics: dict[str, Any], view: SpanTreeView) -> str:
    if span.kind not in {SpanKind.CONVERSATION, SpanKind.LLM_ROUND}:
        return ""
    if span.kind == SpanKind.CONVERSATION:
        dt = view.descendant_tokens.get(span.span_id, _TokenTotals())
        tokens_input, tokens_output = dt.tokens_input, dt.tokens_output
        tokens_cache_read, tokens_reasoning = dt.tokens_cache_read, dt.tokens_reasoning
    else:
        context = f"Span {span.span_id}"
        tokens_input = _detail_int(metrics, TOKENS_INPUT_KEY, context=context, field_name="metrics_json")
        tokens_output = _detail_int(metrics, TOKENS_OUTPUT_KEY, context=context, field_name="metrics_json")
        tokens_cache_read = _detail_int(metrics, TOKENS_CACHE_READ_KEY, context=context, field_name="metrics_json")
        tokens_reasoning = _detail_int(metrics, TOKENS_REASONING_KEY, context=context, field_name="metrics_json")
    parts = [f"{_format_token_count(tokens_input)} in"]
    if tokens_cache_read > 0:
        parts.append(f"{_format_token_count(tokens_cache_read)} cache")
    parts.append(f"{_format_token_count(tokens_output)} out")
    if tokens_reasoning > 0:
        parts.append(f"{_format_token_count(tokens_reasoning)} reasoning")
    return " / ".join(parts)


def _cache_hit_suffix(meta: dict[str, Any]) -> str:
    if _detail_bool(meta, CACHE_HIT_KEY):
        return " cache-hit"
    return ""


def _conversation_label(span: SpanRecord, meta: dict[str, Any]) -> str:
    purpose = _detail_str(meta, PURPOSE_KEY)
    if purpose:
        return purpose
    if span.description:
        return span.description
    return span.name


def _flow_step_label(span: SpanRecord, meta: dict[str, Any]) -> str:
    context = f"Span {span.span_id}"
    step = _detail_int(meta, STEP_KEY, context=context, field_name="meta_json")
    total_steps = _detail_int(meta, TOTAL_STEPS_KEY, context=context, field_name="meta_json")
    if step > 0 and total_steps > 0:
        return f"[{step}/{total_steps}]"
    if step > 0:
        return f"[{step}]"
    if span.sequence_no > 0:
        return f"[{span.sequence_no}]"
    return ""


def _round_label(span: SpanRecord, meta: dict[str, Any]) -> str:
    context = f"Span {span.span_id}"
    round_index = _detail_int(meta, ROUND_INDEX_KEY, context=context, field_name="meta_json")
    if round_index > 0:
        return f"[{round_index}]"
    if span.sequence_no > 0:
        return f"[{span.sequence_no}]"
    return ""


def _span_local_filename(span: SpanRecord, view: SpanTreeView) -> str | None:
    if span.kind in {SpanKind.LLM_ROUND, SpanKind.TOOL_CALL}:
        return None
    if span.kind == SpanKind.DEPLOYMENT and span.parent_span_id is None:
        return "deployment.json"
    if span.parent_span_id is None:
        return None
    siblings = [
        sibling_id
        for sibling_id in view.children_map.get(span.parent_span_id, [])
        if view.spans_by_id[sibling_id].kind not in {SpanKind.LLM_ROUND, SpanKind.TOOL_CALL}
    ]
    sibling_index = siblings.index(span.span_id) + 1
    return span_filename(span.kind, span.name, span.span_id, sibling_index)


def _format_tree_line(span: SpanRecord, view: SpanTreeView, *, depth: int, include_filenames: bool) -> str:
    indent = _FOUR_SPACES * depth
    meta = view.meta_by_id[span.span_id]
    metrics = view.metrics_by_id[span.span_id]
    status = str(span.status)
    duration = _format_duration(span)
    cache_suffix = _cache_hit_suffix(meta)

    if span.kind == SpanKind.CONVERSATION:
        model = _detail_str(meta, MODEL_KEY) or UNKNOWN_MODEL_LABEL
        tokens = _format_token_parts(span, metrics, view)
        cost = span.cost_usd + view.descendant_costs.get(span.span_id, 0.0)
        chain_suffix = f" (continues {str(span.previous_conversation_id)[:8]}…)" if span.previous_conversation_id else ""
        line = f"{indent}conversation: {_conversation_label(span, meta)} {duration} {model}{chain_suffix}"
        if tokens:
            line += f" {tokens}"
        if cost > 0:
            line += f" ${cost:.4f}"
        rendered = f"{line}{cache_suffix}"
        filename = _span_local_filename(span, view) if include_filenames else None
        return f"{rendered}  -> {filename}" if filename is not None else rendered

    if span.kind == SpanKind.LLM_ROUND:
        model = _detail_str(meta, MODEL_KEY) or UNKNOWN_MODEL_LABEL
        tokens = _format_token_parts(span, metrics, view)
        line = f"{indent}llm_round{_round_label(span, meta)}: {model} {duration}"
        if tokens:
            line += f" {tokens}"
        if span.cost_usd > 0:
            line += f" ${span.cost_usd:.4f}"
        rendered = f"{line}{cache_suffix}"
        filename = _span_local_filename(span, view) if include_filenames else None
        return f"{rendered}  -> {filename}" if filename is not None else rendered

    if span.kind == SpanKind.TOOL_CALL:
        tool_name = _detail_str(meta, TOOL_NAME_KEY) or span.name
        return f"{indent}tool_call{_round_label(span, meta)}: {tool_name} {status} {duration}{cache_suffix}"

    if span.kind == SpanKind.FLOW:
        step_label = _flow_step_label(span, meta)
        rendered = f"{indent}flow{step_label}: {span.name} {status} {duration}{cache_suffix}"
        filename = _span_local_filename(span, view) if include_filenames else None
        return f"{rendered}  -> {filename}" if filename is not None else rendered

    rendered = f"{indent}{span.kind}: {span.name} {status} {duration}{cache_suffix}"
    filename = _span_local_filename(span, view) if include_filenames else None
    return f"{rendered}  -> {filename}" if filename is not None else rendered


def format_span_tree_lines(view: SpanTreeView, *, include_filenames: bool = False) -> list[str]:
    """Render the full span execution tree as indented plain-text lines."""
    lines: list[str] = []
    visited: set[UUID] = set()

    def append_tree_lines(span_id: UUID, depth: int) -> None:
        if span_id in visited:
            lines.append(f"{_FOUR_SPACES * depth}[cycle detected while rendering execution tree]")
            return
        if depth > MAX_TREE_DEPTH:
            lines.append(f"{_FOUR_SPACES * depth}[execution tree depth limit reached]")
            return

        visited.add(span_id)
        span = view.spans_by_id[span_id]
        lines.append(_format_tree_line(span, view, depth=depth, include_filenames=include_filenames))
        for child_id in view.children_map.get(span_id, []):
            append_tree_lines(child_id, depth + 1)
        visited.remove(span_id)

    append_tree_lines(view.root_span.span_id, 0)
    return lines


def format_span_overview_lines(view: SpanTreeView) -> list[str]:
    """Render a compact plain-text deployment overview for ai-trace show."""
    total_tokens = view.totals.tokens_input + view.totals.tokens_output + view.totals.tokens_cache_read + view.totals.tokens_reasoning
    return [
        f"Deployment {view.root_span.deployment_name or view.root_span.name} / {view.root_span.run_id}",
        (
            f"Status: {view.root_span.status}  Duration: {_format_duration(view.root_span)}  "
            f"Spans: {sum(view.counts_by_kind.values())}  Documents: {len(view.document_shas)}"
        ),
        (
            f"Flows: {view.counts_by_kind.get(SpanKind.FLOW, 0)}  "
            f"Tasks: {view.counts_by_kind.get(SpanKind.TASK, 0)}  "
            f"Operations: {view.counts_by_kind.get(SpanKind.OPERATION, 0)}  "
            f"Conversations: {view.counts_by_kind.get(SpanKind.CONVERSATION, 0)}  "
            f"LLM Rounds: {view.counts_by_kind.get(SpanKind.LLM_ROUND, 0)}  "
            f"Tool Calls: {view.counts_by_kind.get(SpanKind.TOOL_CALL, 0)}"
        ),
        f"Total Tokens: {total_tokens:,}  Total Cost: ${view.totals.cost_usd:.4f}",
    ]


def _build_failures_lines(view: SpanTreeView) -> list[str]:
    failed_spans = sorted(
        (span for span in view.spans_by_id.values() if span.status == SpanStatus.FAILED),
        key=lambda span: (span.started_at, span.sequence_no, str(span.span_id)),
    )
    if not failed_spans:
        return []

    lines = ["## Failures", ""]
    for span in failed_spans:
        error_parts = [part for part in (span.error_type, span.error_message) if part]
        error_text = ": ".join(error_parts) if error_parts else "(no error recorded)"
        lines.append(f"- `{span.kind}: {span.name}`")
        lines.append(f"  `{error_text}`")
    lines.append("")
    return lines


def _build_flow_plan_lines(view: SpanTreeView) -> list[str]:
    meta = view.meta_by_id[view.root_span.span_id]
    flow_plan = meta.get(FLOW_PLAN_KEY)
    if not isinstance(flow_plan, list) or not flow_plan:
        return []

    flows_by_name: dict[str, list[SpanRecord]] = {}
    for span in view.spans_by_id.values():
        if span.kind != SpanKind.FLOW:
            continue
        flows_by_name.setdefault(span.name, []).append(span)

    for spans in flows_by_name.values():
        spans.sort(key=lambda span: (span.sequence_no, span.started_at, str(span.span_id)))

    lines = ["## Flow Plan", "", "| Step | Flow | Status | Duration | Cost |", "|---|---|---|---:|---:|"]
    planned_flow_span_ids: set[UUID] = set()
    for index, plan_entry in enumerate(flow_plan, start=1):
        if not isinstance(plan_entry, dict):
            msg = (
                f"Span {view.root_span.span_id} meta_json['flow_plan'] must contain only objects. "
                "Persist flow_plan as a JSON array of objects with at least a name field."
            )
            raise TypeError(msg)
        name = plan_entry.get("name")
        if not isinstance(name, str) or not name:
            name = f"Flow {index}"
        matches = flows_by_name.get(name, [])
        flow = matches.pop(0) if matches else None
        if flow is None:
            lines.append(f"| {index} | {name} | skipped | - | $0.0000 |")
            continue
        planned_flow_span_ids.add(flow.span_id)
        cost = flow.cost_usd + view.descendant_costs.get(flow.span_id, 0.0)
        lines.append(f"| {index} | {name} | {flow.status} | {_format_duration(flow)} | ${cost:.4f} |")

    unmatched_flows = sorted(
        (span for span in view.spans_by_id.values() if span.kind == SpanKind.FLOW and span.span_id not in planned_flow_span_ids),
        key=lambda span: (span.sequence_no, span.started_at, str(span.span_id)),
    )
    for flow in unmatched_flows:
        cost = flow.cost_usd + view.descendant_costs.get(flow.span_id, 0.0)
        lines.append(f"| {flow.sequence_no or '-'} | {flow.name} | {flow.status} | {_format_duration(flow)} | ${cost:.4f} |")
    lines.append("")
    return lines


def _build_conversation_chain_lines(view: SpanTreeView) -> list[str]:
    """Render conversation chains showing warmup → fork topology."""
    conversation_spans = [span for span in view.spans_by_id.values() if span.kind == SpanKind.CONVERSATION and span.previous_conversation_id is not None]
    if not conversation_spans:
        return []
    children_of: dict[UUID, list[SpanRecord]] = {}
    for span in conversation_spans:
        assert span.previous_conversation_id is not None
        children_of.setdefault(span.previous_conversation_id, []).append(span)
    roots: set[UUID] = set()
    for parent_id in children_of:
        parent = view.spans_by_id.get(parent_id)
        if parent is not None and parent.previous_conversation_id is None:
            roots.add(parent_id)
    if not roots:
        return []
    lines = ["## Conversation Chains", ""]
    for root_id in sorted(roots, key=lambda rid: view.spans_by_id[rid].started_at):
        root = view.spans_by_id[root_id]
        forks = sorted(children_of.get(root_id, []), key=lambda s: s.started_at)
        fork_names = ", ".join(s.name for s in forks)
        lines.append(f"- {root.name} → [{fork_names}]")
    lines.append("")
    return lines


def generate_summary_from_tree(tree: list[SpanRecord], root_deployment_id: UUID) -> str:
    """Render summary.md content for a span-era deployment tree."""
    view = build_span_tree_view(tree, root_deployment_id)
    if view is None:
        return "# No execution data found\n"

    total_tokens = view.totals.tokens_input + view.totals.tokens_output + view.totals.tokens_cache_read + view.totals.tokens_reasoning
    header = f"# {view.root_span.deployment_name or view.root_span.name} / {view.root_span.run_id}"
    lines = [header]
    if view.root_span.description:
        lines.append(f"\n> {view.root_span.description}")
    lines.extend([
        "",
        (
            f"**Status**: {view.root_span.status} | **Duration**: {_format_duration(view.root_span)} | "
            f"**Spans**: {sum(view.counts_by_kind.values())} | **Documents**: {len(view.document_shas)}"
        ),
        (
            f"**Flows**: {view.counts_by_kind.get(SpanKind.FLOW, 0)} | "
            f"**Tasks**: {view.counts_by_kind.get(SpanKind.TASK, 0)} | "
            f"**Operations**: {view.counts_by_kind.get(SpanKind.OPERATION, 0)} | "
            f"**Conversations**: {view.counts_by_kind.get(SpanKind.CONVERSATION, 0)} | "
            f"**LLM Rounds**: {view.counts_by_kind.get(SpanKind.LLM_ROUND, 0)} | "
            f"**Tool Calls**: {view.counts_by_kind.get(SpanKind.TOOL_CALL, 0)}"
        ),
        f"**Total Tokens**: {total_tokens:,} | **Total Cost**: ${view.totals.cost_usd:.4f}",
        "",
        "## Navigation",
        "",
        f"- `runs/{run_directory_name(view.root_span.started_at, view.root_span.name, view.root_span.span_id)}/` — hierarchical execution tree",
        "- `documents/` — content-addressed document metadata grouped by type",
        "- `blobs/` — flat binary blob storage",
        "- `logs.jsonl` — chronological execution logs",
        "- `costs.md` — cost breakdown by model",
        "- `llm_calls.jsonl` — chronological LLM round index",
        "- `errors.md` — failed span report when failures exist",
        "- `documents.md` — document catalog when documents exist",
        "",
        "## CLI Commands",
        "",
        "```bash",
        f"ai-trace show {root_deployment_id} --db-path <download-dir>",
        "",
        "# Replay a specific span",
        "ai-replay show --from-db <span-id> --db-path <download-dir>",
        "ai-replay run --from-db <span-id> --db-path <download-dir>",
        "```",
        "",
    ])
    lines.extend(_build_flow_plan_lines(view))
    lines.extend(_build_conversation_chain_lines(view))
    lines.extend(_build_failures_lines(view))
    lines.extend(["## Execution Tree", ""])
    lines.extend(format_span_tree_lines(view, include_filenames=True))
    lines.append("")
    return "\n".join(lines)


def generate_costs_from_tree(tree: list[SpanRecord], root_deployment_id: UUID) -> str:
    """Render costs.md content for a span-era deployment tree."""
    view = build_span_tree_view(tree, root_deployment_id)
    if view is None:
        return ""

    per_model: dict[str, CostTotals] = {}
    rounds_per_model: Counter[str] = Counter()
    for span in tree:
        if span.kind != SpanKind.LLM_ROUND:
            continue
        meta = view.meta_by_id[span.span_id]
        metrics = view.metrics_by_id[span.span_id]
        context = f"Span {span.span_id}"
        model = _detail_str(meta, MODEL_KEY) or UNKNOWN_MODEL_LABEL
        current = per_model.get(model, CostTotals())
        per_model[model] = CostTotals(
            cost_usd=current.cost_usd + span.cost_usd,
            tokens_input=current.tokens_input + _detail_int(metrics, TOKENS_INPUT_KEY, context=context, field_name="metrics_json"),
            tokens_output=current.tokens_output + _detail_int(metrics, TOKENS_OUTPUT_KEY, context=context, field_name="metrics_json"),
            tokens_cache_read=current.tokens_cache_read + _detail_int(metrics, TOKENS_CACHE_READ_KEY, context=context, field_name="metrics_json"),
            tokens_reasoning=current.tokens_reasoning + _detail_int(metrics, TOKENS_REASONING_KEY, context=context, field_name="metrics_json"),
        )
        rounds_per_model[model] += 1

    lines = ["# Cost by Model", ""]
    if per_model:
        ordered_models = sorted(
            per_model.items(),
            key=lambda item: (-item[1].cost_usd, item[0]),
        )
        lines.extend(["| Model | Rounds | Input | Cache Read | Output | Reasoning | Cost |", "|---|---:|---:|---:|---:|---:|---:|"])
        for model, totals in ordered_models:
            lines.append(
                f"| {model} | {rounds_per_model[model]} | {totals.tokens_input:,} | {totals.tokens_cache_read:,} | "
                f"{totals.tokens_output:,} | {totals.tokens_reasoning:,} | ${totals.cost_usd:.4f} |"
            )
        lines.append("")
    lines.extend([
        f"**Total**: ${view.totals.cost_usd:.4f}",
        "",
    ])
    return "\n".join(lines)



# pyright: reportPrivateUsage=false
"""Tests for task retries, timeouts, cancellation, and retry span detail."""

import asyncio
import json
from types import MappingProxyType
from uuid import uuid7

import pytest

from ai_pipeline_core._llm_core.model_response import ModelResponse
from ai_pipeline_core._llm_core.types import TokenUsage
from ai_pipeline_core.database import SpanKind, SpanStatus
from ai_pipeline_core.database._memory import _MemoryDatabase
from ai_pipeline_core.deployment._types import _NoopPublisher
from ai_pipeline_core.documents import Document
from ai_pipeline_core.llm.conversation import Conversation
from ai_pipeline_core.pipeline import PipelineTask, pipeline_test_context
from ai_pipeline_core.pipeline._execution_context import ExecutionContext, FlowFrame, set_execution_context
from ai_pipeline_core.pipeline._runtime_sinks import build_runtime_sinks
from ai_pipeline_core.pipeline.limits import _SharedStatus
from ai_pipeline_core.settings import settings


class EdgeInputDoc(Document):
    """Input for resilience tests."""


class EdgeOutputDoc(Document):
    """Output for resilience tests."""


def _make_response(
    *,
    content: str,
    prompt_tokens: int,
    completion_tokens: int,
    cost: float,
    response_id: str,
) -> ModelResponse[str]:
    return ModelResponse[str](
        content=content,
        parsed=content,
        usage=TokenUsage(
            prompt_tokens=prompt_tokens,
            completion_tokens=completion_tokens,
            total_tokens=prompt_tokens + completion_tokens,
        ),
        cost=cost,
        model="test-model",
        response_id=response_id,
        metadata={"time_taken": 0.2, "first_token_time": 0.1},
        tool_calls=(),
    )


def _make_fake_generate(results: list[ModelResponse[str] | BaseException]):
    remaining = list(results)

    async def _fake_generate(*args: object, **kwargs: object) -> ModelResponse[str]:
        _ = (args, kwargs)
        next_item = remaining.pop(0)
        if isinstance(next_item, BaseException):
            raise next_item
        return next_item

    return _fake_generate


class _CancelOnFirstAttempt(PipelineTask):
    retries = 2
    retry_delay_seconds = 0
    attempt_count = 0

    @classmethod
    async def run(cls, documents: tuple[EdgeInputDoc, ...]) -> tuple[EdgeOutputDoc, ...]:
        _ = documents
        cls.attempt_count += 1
        raise asyncio.CancelledError("externally cancelled")


class _TimeoutWithRetry(PipelineTask):
    retries = 1
    retry_delay_seconds = 0
    timeout_seconds = 1
    attempt_count = 0

    @classmethod
    async def run(cls, documents: tuple[EdgeInputDoc, ...]) -> tuple[EdgeOutputDoc, ...]:
        cls.attempt_count += 1
        if cls.attempt_count == 1:
            await asyncio.sleep(10)
        return (EdgeOutputDoc.derive(derived_from=(documents[0],), name="out.txt", content="ok"),)


class _InstantTask(PipelineTask):
    @classmethod
    async def run(cls, documents: tuple[EdgeInputDoc, ...]) -> tuple[EdgeOutputDoc, ...]:
        return (EdgeOutputDoc.derive(derived_from=(documents[0],), name="instant.txt", content="done"),)


class _RetryWithConversationTask(PipelineTask):
    retries = 1
    retry_delay_seconds = 0
    attempt_count = 0

    @classmethod
    async def run(cls, documents: tuple[EdgeInputDoc, ...]) -> tuple[EdgeOutputDoc, ...]:
        cls.attempt_count += 1
        conv = Conversation(model="test-model", enable_substitutor=False)
        conv = await conv.send(f"attempt-{cls.attempt_count}", purpose=f"attempt-{cls.attempt_count}")
        if cls.attempt_count == 1:
            raise RuntimeError("retry me")
        return (EdgeOutputDoc.derive(derived_from=(documents[0],), name="retried.txt", content=conv.content),)


def _make_input() -> EdgeInputDoc:
    return EdgeInputDoc.create_root(name="in.txt", content="x", reason="task-resilience-test")


def _make_retry_test_context(database: _MemoryDatabase) -> ExecutionContext:
    deployment_id = uuid7()
    flow_span_id = uuid7()
    return ExecutionContext(
        run_id="test-run",
        execution_id=None,
        publisher=_NoopPublisher(),
        limits=MappingProxyType({}),
        limits_status=_SharedStatus(),
        database=database,
        sinks=build_runtime_sinks(database=database, settings_obj=settings),
        deployment_id=deployment_id,
        root_deployment_id=deployment_id,
        deployment_name="test-pipeline",
        flow_frame=FlowFrame(
            name="test-flow",
            flow_class_name="TestFlow",
            step=1,
            total_steps=1,
            flow_minutes=(1.0,),
            completed_minutes=0.0,
            flow_params={},
        ),
        span_id=flow_span_id,
        current_span_id=flow_span_id,
        flow_span_id=flow_span_id,
    )


@pytest.mark.asyncio
async def test_cancelled_error_bypasses_retries() -> None:
    _CancelOnFirstAttempt.attempt_count = 0
    with pipeline_test_context():
        with pytest.raises(asyncio.CancelledError):
            await _CancelOnFirstAttempt.run((_make_input(),))

    assert _CancelOnFirstAttempt.attempt_count == 1


@pytest.mark.asyncio
async def test_timeout_triggers_retry_then_succeeds() -> None:
    _TimeoutWithRetry.attempt_count = 0
    with pipeline_test_context():
        result = await _TimeoutWithRetry.run((_make_input(),))

    assert _TimeoutWithRetry.attempt_count == 2
    assert result[0].name == "out.txt"


@pytest.mark.asyncio
async def test_handle_can_be_awaited_twice() -> None:
    with pipeline_test_context():
        handle = _InstantTask.run((_make_input(),))
        first = await handle
        second = await handle

    assert first[0].sha256 == second[0].sha256


@pytest.mark.asyncio
async def test_handle_concurrent_await_returns_same_result() -> None:
    with pipeline_test_context():
        handle = _InstantTask.run((_make_input(),))
        first, second = await asyncio.gather(handle.result(), handle.result())

    assert first[0].sha256 == second[0].sha256


@pytest.mark.asyncio
async def test_retried_task_persists_final_attempt_and_failed_attempt_llm_rounds(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setattr(
        "ai_pipeline_core.llm.conversation.core_generate",
        _make_fake_generate([
            _make_response(content="first", prompt_tokens=10, completion_tokens=5, cost=0.3, response_id="resp-1"),
            _make_response(content="second", prompt_tokens=20, completion_tokens=6, cost=0.7, response_id="resp-2"),
        ]),
    )
    _RetryWithConversationTask.attempt_count = 0
    database = _MemoryDatabase()
    with set_execution_context(_make_retry_test_context(database)):
        result = await _RetryWithConversationTask.run((_make_input(),))

    assert result[0].name == "retried.txt"
    task_span = next(span for span in database._spans.values() if span.kind == SpanKind.TASK)
    task_meta = json.loads(task_span.meta_json)
    llm_rounds = [span for span in database._spans.values() if span.kind == SpanKind.LLM_ROUND]

    assert task_span.status == SpanStatus.COMPLETED
    assert task_meta["attempt"] == 1
    assert len(llm_rounds) == 2
    assert sorted(span.cost_usd for span in llm_rounds) == [0.3, 0.7]


class _RetryLeakTask(PipelineTask):
    retries = 1
    retry_delay_seconds = 0
    attempt_count = 0

    @classmethod
    async def run(cls, documents: tuple[EdgeInputDoc, ...]) -> tuple[EdgeOutputDoc, ...]:
        cls.attempt_count += 1
        result = EdgeOutputDoc.derive(derived_from=documents, name=f"attempt-{cls.attempt_count}.txt", content="ok")
        if cls.attempt_count == 1:
            raise RuntimeError("retry once")
        return (result,)


@pytest.mark.asyncio
async def test_retry_discards_created_documents_from_failed_attempts(caplog: pytest.LogCaptureFixture) -> None:
    _RetryLeakTask.attempt_count = 0
    caplog.set_level("WARNING", logger="ai_pipeline_core.pipeline._task")
    with pipeline_test_context():
        result = await _RetryLeakTask.run((_make_input(),))

    assert result[0].name == "attempt-2.txt"
    assert "Orphaned document" not in caplog.text



"""Tests for conversation message serialization helpers."""

import base64

from ai_pipeline_core._llm_core import CoreMessage, Role
from ai_pipeline_core._llm_core.types import ImageContent, PDFContent, TextContent
from ai_pipeline_core.documents._hashing import compute_content_sha256


def test_core_messages_to_db_span_input_uses_doc_ref() -> None:
    from ai_pipeline_core.llm._conversation_messages import _core_messages_to_db_span_input

    image_bytes = b"\x89PNG\r\n\x1a\n" + (b"\x00" * 32)
    pdf_bytes = b"%PDF-1.7\n" + (b"0" * 32)
    messages = [
        CoreMessage(
            role=Role.USER,
            content=(
                TextContent(text="before"),
                ImageContent(data=base64.b64encode(image_bytes), mime_type="image/png"),
                PDFContent(data=base64.b64encode(pdf_bytes)),
            ),
        )
    ]

    result = _core_messages_to_db_span_input(messages)

    assert result == [
        {
            "role": "user",
            "content": [
                {"type": "text", "text": "before"},
                {"type": "image", "$doc_ref": compute_content_sha256(image_bytes), "media_type": "image/png"},
                {"type": "pdf", "$doc_ref": compute_content_sha256(pdf_bytes), "media_type": "application/pdf"},
            ],
        }
    ]



"""Tool execution loop for Conversation.

Implements the auto-loop that calls the LLM, executes requested tools, and
re-sends results until the LLM produces a final answer or max rounds is reached.
"""

import asyncio
import json
from collections.abc import Callable, Coroutine
from typing import Any

from pydantic import BaseModel, ValidationError

from ai_pipeline_core._llm_core.model_response import ModelResponse
from ai_pipeline_core._llm_core.types import CoreMessage, ModelOptions, RawToolCall, Role, TokenUsage
from ai_pipeline_core.database import SpanKind
from ai_pipeline_core.logger import get_pipeline_logger
from ai_pipeline_core.pipeline._execution_context import get_execution_context, get_sinks
from ai_pipeline_core.pipeline._track_span import track_span

from ._substitutor import URLSubstitutor
from .tools import Tool, ToolCallRecord, ToolOutput

__all__: list[str] = []

logger = get_pipeline_logger(__name__)


async def _execute_single_tool(
    tool: Tool,
    tool_call: RawToolCall,
    round_num: int,
) -> tuple[ToolCallRecord | None, ToolOutput]:
    """Execute a single tool call with error handling and timeout."""
    tool_cls = type(tool)
    snake_name = tool_cls.name
    execution_ctx = get_execution_context()
    tool_target = f"tool_call:{tool_cls.__module__}:{tool_cls.__qualname__}"
    receiver_payload = {
        "mode": "tool_ref",
        "value": {"name": tool_cls.name, "class_path": f"{tool_cls.__module__}:{tool_cls.__qualname__}"},
    }

    async with track_span(
        SpanKind.TOOL_CALL,
        snake_name,
        tool_target,
        sinks=get_sinks(),
        encode_receiver=receiver_payload,
        encode_input={"input": _parse_tool_arguments(tool_call.arguments), "tool_call_id": tool_call.id, "round_index": round_num},
        db=execution_ctx.database if execution_ctx is not None else None,
        input_preview={"tool_name": snake_name, "arguments": _parse_tool_arguments(tool_call.arguments)},
    ) as span_ctx:
        try:
            parsed_input = tool_cls.Input.model_validate_json(tool_call.arguments)
        except (ValidationError, json.JSONDecodeError) as e:
            logger.warning("Tool input validation failed for %s: %s", tool_cls.__name__, e)
            output = ToolOutput(content=f"Error: Invalid arguments for tool '{snake_name}': {e}")
            span_ctx.set_meta(
                tool_name=snake_name,
                tool_class_path=f"{tool_cls.__module__}:{tool_cls.__qualname__}",
                tool_call_id=tool_call.id,
                round_index=round_num,
            )
            span_ctx.set_output_preview(output.model_dump(mode="json"))
            span_ctx._set_output_value(output)
            return None, output
        try:
            result = await tool.execute(parsed_input)
        except TypeError, AssertionError:
            raise
        except Exception as e:
            logger.warning("Tool execution failed for %s: %s", tool_cls.__name__, e)
            output = ToolOutput(content=f"Error: Tool '{snake_name}' failed: {e}")
            span_ctx.set_meta(
                tool_name=snake_name,
                tool_class_path=f"{tool_cls.__module__}:{tool_cls.__qualname__}",
                tool_call_id=tool_call.id,
                round_index=round_num,
            )
            span_ctx.set_output_preview(output.model_dump(mode="json"))
            span_ctx._set_output_value(output)
            return (ToolCallRecord(tool=tool_cls, input=parsed_input, output=output, round=round_num), output)

        record = ToolCallRecord(tool=tool_cls, input=parsed_input, output=result, round=round_num)
        span_ctx.set_meta(
            tool_name=snake_name,
            tool_class_path=f"{tool_cls.__module__}:{tool_cls.__qualname__}",
            tool_call_id=tool_call.id,
            round_index=round_num,
        )
        span_ctx.set_output_preview(result.model_dump(mode="json"))
        span_ctx._set_output_value(result)
        return record, result


InvokeLLMFn = Callable[..., Coroutine[Any, Any, ModelResponse[Any]]]


async def _execute_all_tool_calls(
    tool_calls: tuple[RawToolCall, ...],
    tool_lookup: dict[str, Tool],
    round_num: int,
) -> list[tuple[RawToolCall, ToolCallRecord | None, ToolOutput]]:
    """Execute all tool calls in parallel, returning results in original order."""

    async def _execute_one(tc: RawToolCall) -> tuple[RawToolCall, ToolCallRecord | None, ToolOutput]:
        if tc.function_name not in tool_lookup:
            available = ", ".join(sorted(tool_lookup.keys()))
            output = ToolOutput(content=f"Error: Unknown tool '{tc.function_name}'. Available tools: {available}")
            return tc, None, output
        record, output = await _execute_single_tool(tool_lookup[tc.function_name], tc, round_num)
        return tc, record, output

    results = await asyncio.gather(*(_execute_one(tc) for tc in tool_calls), return_exceptions=True)
    merged: list[tuple[RawToolCall, ToolCallRecord | None, ToolOutput]] = []
    for i, result in enumerate(results):
        if isinstance(result, BaseException):
            if isinstance(result, (TypeError, AssertionError, KeyboardInterrupt, SystemExit, asyncio.CancelledError)):
                raise result
            logger.warning("Unexpected error executing tool: %s", result)
            output = ToolOutput(content=f"Error: {result}")
            merged.append((tool_calls[i], None, output))
        else:
            merged.append(result)
    return merged


async def execute_tool_loop(
    *,
    invoke_llm: InvokeLLMFn,
    tool_schemas: list[dict[str, Any]],
    tool_lookup: dict[str, Tool],
    tool_choice: str | None,
    max_tool_rounds: int,
    purpose: str | None,
    expected_cost: float | None,
    core_messages: list[CoreMessage],
    context_count: int,
    effective_options: ModelOptions | None,
    substitutor: URLSubstitutor | None,
    build_tool_result_message: Callable[[str, str, str], Any],
    response_format: type[BaseModel] | None = None,
) -> tuple[list[Any], ModelResponse[Any], tuple[ToolCallRecord, ...]]:
    """Execute the tool auto-loop.

    Returns (accumulated_new_messages, final_response, tool_call_records).
    """
    if max_tool_rounds < 1:
        raise ValueError(f"max_tool_rounds must be >= 1, got {max_tool_rounds}")

    accumulated_messages: list[Any] = []
    all_records: list[ToolCallRecord] = []

    for round_num in range(1, max_tool_rounds + 1):
        logger.info("Tool loop round %d/%d (purpose=%s)", round_num, max_tool_rounds, purpose or "unspecified")
        response: ModelResponse[Any] = await invoke_llm(
            core_messages=core_messages,
            effective_options=effective_options,
            context_count=context_count,
            tools=tool_schemas,
            tool_choice=tool_choice if round_num == 1 else None,
            response_format=response_format,
            purpose=f"{purpose}:tool_round_{round_num}" if purpose else f"tool_round_{round_num}",
            expected_cost=expected_cost,
            round_index=round_num,
            tool_schemas=tool_schemas,
        )

        if not response.has_tool_calls:
            accumulated_messages.append(response)
            return accumulated_messages, response, tuple(all_records)

        accumulated_messages.append(response)
        core_messages.append(CoreMessage(role=Role.ASSISTANT, content=response.content or "", tool_calls=response.tool_calls))

        tool_names = ", ".join(tc.function_name for tc in response.tool_calls)
        logger.info("Tool loop round %d: executing %d tool(s): %s", round_num, len(response.tool_calls), tool_names)
        merged = await _execute_all_tool_calls(response.tool_calls, tool_lookup, round_num)

        for tc, record, output in merged:
            if record:
                all_records.append(record)
            result_content = substitutor.substitute(output.content) if substitutor else output.content
            accumulated_messages.append(build_tool_result_message(tc.id, tc.function_name, output.content))
            core_messages.append(CoreMessage(role=Role.TOOL, content=result_content, tool_call_id=tc.id, name=tc.function_name))

    # max_tool_rounds exhausted — force a final text response without tools
    logger.warning("Tool loop reached max_tool_rounds=%d. Forcing final response without tools.", max_tool_rounds)
    try:
        response = await invoke_llm(
            core_messages=core_messages,
            effective_options=effective_options,
            context_count=context_count,
            tools=[],
            tool_choice=None,
            response_format=response_format,
            purpose=f"{purpose}:forced_final" if purpose else "forced_final",
            expected_cost=expected_cost,
            round_index=max_tool_rounds + 1,
            tool_schemas=[],
        )
    except Exception:
        logger.warning(
            "Forced final response failed after max_tool_rounds=%d. Returning accumulated tool results (%d records) without final synthesis.",
            max_tool_rounds,
            len(all_records),
        )
        response = ModelResponse[Any](
            content="[Final synthesis failed after max tool rounds. Tool results are preserved in tool_call_records.]",
            parsed="[Final synthesis failed after max tool rounds.]",
            usage=TokenUsage(prompt_tokens=0, completion_tokens=0, total_tokens=0),
            model="",
        )
    accumulated_messages.append(response)
    return accumulated_messages, response, tuple(all_records)


def _parse_tool_arguments(arguments: str) -> Any:
    """Best-effort parse of tool call arguments for tracing payloads."""
    try:
        return json.loads(arguments)
    except json.JSONDecodeError:
        return {"_raw": arguments}



"""Integration tests for tool calling across providers.

Tests the complete tool loop with real LLM APIs: grok-4.1-fast, gemini-3-flash, gpt-5-mini.
Each test is parameterized by model AND streaming mode to verify both paths.
"""

import pytest
from pydantic import BaseModel, Field

from ai_pipeline_core import Conversation, ModelOptions, Tool
from ai_pipeline_core.settings import settings

HAS_API_KEYS = bool(settings.openai_api_key and settings.openai_base_url)


class CityInfo(BaseModel):
    city: str = Field(description="Capital city name")
    country: str = Field(description="Country name")


pytestmark = [
    pytest.mark.integration,
    pytest.mark.skipif(not HAS_API_KEYS, reason="OpenAI API keys not configured"),
]

# Models to test tool calling
TOOL_MODELS = ("grok-4.1-fast", "gemini-3-flash", "gpt-5-mini")
STREAM_MODES = (False, True)


@pytest.fixture(params=[pytest.param((model, stream), id=f"{model}-{'stream' if stream else 'nostream'}") for model in TOOL_MODELS for stream in STREAM_MODES])
def model_opts(request: pytest.FixtureRequest) -> tuple[str, ModelOptions]:
    """Parameterized (model_name, ModelOptions) fixture for all model x stream combos."""
    model, stream = request.param
    return model, ModelOptions(stream=stream)


# ── Test Tools ───────────────────────────────────────────────────────────────


class Calculator(Tool):
    """Perform basic arithmetic calculations."""

    class Input(BaseModel):
        expression: str = Field(description="Mathematical expression to evaluate, e.g. '2 + 3'")

    class Output(BaseModel):
        result: str

    async def run(self, input: Input) -> Output:
        try:
            result = eval(input.expression, {"__builtins__": {}}, {})
            return self.Output(result=str(result))
        except (SyntaxError, TypeError, ValueError, ZeroDivisionError, NameError, ArithmeticError) as e:
            return self.Output(result=f"Error: {e}")


class GetCapital(Tool):
    """Get the capital city of a country."""

    class Input(BaseModel):
        country: str = Field(description="Country name")

    class Output(BaseModel):
        answer: str

    async def run(self, input: Input) -> Output:
        capitals = {
            "france": "Paris",
            "japan": "Tokyo",
            "brazil": "Brasília",
            "germany": "Berlin",
        }
        country = input.country.lower()
        if country in capitals:
            return self.Output(answer=f"The capital of {input.country} is {capitals[country]}")
        return self.Output(answer=f"Unknown country: {input.country}")


class FailingTool(Tool):
    """A tool that always fails to test error handling."""

    class Input(BaseModel):
        reason: str = Field(description="Reason for failure")

    class Output(BaseModel):
        result: str

    async def run(self, input: Input) -> Output:
        raise RuntimeError(f"Intentional failure: {input.reason}")


class WeatherLookup(Tool):
    """Look up weather for a city."""

    class Input(BaseModel):
        city: str = Field(description="City to look up weather for")

    class Output(BaseModel):
        weather: str

    async def run(self, input: Input) -> Output:
        return self.Output(weather=f"The weather in {input.city} is sunny, 22°C")


# ── Integration Tests ────────────────────────────────────────────────────────


async def test_single_tool_call(model_opts: tuple[str, ModelOptions]) -> None:
    """LLM calls a single tool and uses the result in its response."""
    model, opts = model_opts
    conv = await Conversation(model=model, model_options=opts).send(
        "What is the capital of France? Use the get_capital tool.",
        tools=[GetCapital()],
        purpose="test_single_tool",
    )

    assert "Paris" in conv.content
    assert len(conv.tool_call_records) >= 1
    assert conv.tool_call_records[0].tool is GetCapital


async def test_tool_with_calculation(model_opts: tuple[str, ModelOptions]) -> None:
    """LLM uses a calculator tool."""
    model, opts = model_opts
    conv = await Conversation(model=model, model_options=opts).send(
        "What is 137 * 29? Use the calculator tool to compute it.",
        tools=[Calculator()],
        purpose="test_calculator",
    )

    # Some models format with commas (3,973) or spaces
    assert "3973" in conv.content.replace(",", "").replace(" ", "")
    assert len(conv.tool_call_records) >= 1


async def test_multiple_tools_available(model_opts: tuple[str, ModelOptions]) -> None:
    """LLM correctly selects from multiple available tools."""
    model, opts = model_opts
    conv = await Conversation(model=model, model_options=opts).send(
        "What is the capital of Japan? Use the appropriate tool.",
        tools=[Calculator(), GetCapital()],
        purpose="test_multi_tool_selection",
    )

    assert "Tokyo" in conv.content
    assert any(r.tool is GetCapital for r in conv.tool_call_records)


async def test_tool_error_recovery(model_opts: tuple[str, ModelOptions]) -> None:
    """LLM receives tool error and gracefully handles it."""
    model, opts = model_opts
    conv = await Conversation(model=model, model_options=opts).send(
        "Use the failing_tool with reason 'test'. Then tell me what happened.",
        tools=[FailingTool()],
        max_tool_rounds=2,
        purpose="test_error_recovery",
    )

    # LLM should have received the error and responded about it
    assert conv.content  # non-empty response
    assert len(conv.messages) > 1  # at least user + error response + final


async def test_tool_choice_none(model_opts: tuple[str, ModelOptions]) -> None:
    """tool_choice='none' prevents tool usage."""
    model, opts = model_opts
    conv = await Conversation(model=model, model_options=opts).send(
        "What is the capital of Germany?",
        tools=[GetCapital()],
        tool_choice="none",
        purpose="test_tool_choice_none",
    )

    # Should have a response without any tool calls
    assert conv.content
    assert len(conv.tool_call_records) == 0


async def test_max_tool_rounds_forced_final(model_opts: tuple[str, ModelOptions]) -> None:
    """max_tool_rounds=1 forces a final response after one round."""
    model, opts = model_opts
    conv = await Conversation(model=model, model_options=opts).send(
        "What is the capital of France? Use the get_capital tool.",
        tools=[GetCapital()],
        max_tool_rounds=1,
        purpose="test_max_rounds",
    )

    # Should still get a response (forced final or natural completion)
    assert conv.content


async def test_structured_output_with_tools(model_opts: tuple[str, ModelOptions]) -> None:
    """Tools work with structured output (tool loop → final structured call)."""

    model, opts = model_opts
    conv = await Conversation(model=model, model_options=opts).send_structured(
        "Look up the capital of France using the tool, then return the result.",
        response_format=CityInfo,
        tools=[GetCapital()],
        purpose="test_structured_with_tools",
    )

    assert conv.parsed is not None
    result = conv.parsed
    assert isinstance(result, CityInfo)
    assert result.city.lower() == "paris"


async def test_multi_round_tool_conversation(model_opts: tuple[str, ModelOptions]) -> None:
    """Multi-round tool conversation within a single send call."""
    model, opts = model_opts
    # Use two tools that the LLM might call in sequence
    conv = await Conversation(model=model, model_options=opts).send(
        "First look up the weather in Paris using the weather_lookup tool, then calculate 22 * 3 using the calculator tool. Report both results.",
        tools=[WeatherLookup(), Calculator()],
        purpose="test_multi_round",
    )

    assert conv.content
    # Should have used at least one tool
    assert len(conv.tool_call_records) >= 1


async def test_tool_choice_required(model_opts: tuple[str, ModelOptions]) -> None:
    """tool_choice='required' forces at least one tool call."""
    model, opts = model_opts
    conv = await Conversation(model=model, model_options=opts).send(
        "Tell me about the weather.",
        tools=[WeatherLookup()],
        tool_choice="required",
        purpose="test_tool_choice_required",
    )

    assert conv.content
    assert len(conv.tool_call_records) >= 1


async def test_conversation_continues_after_tools() -> None:
    """Conversation state is preserved after tool usage for follow-up messages."""
    conv = await Conversation(
        model="gemini-3-flash",
        model_options=ModelOptions(stream=False),
    ).send(
        "What is the capital of France? Use the get_capital tool.",
        tools=[GetCapital()],
        purpose="test_continuation_1",
    )

    assert "Paris" in conv.content

    # Follow-up without tools should work and have conversation history
    conv = await conv.send(
        "What country was that capital in?",
        purpose="test_continuation_2",
    )

    assert "France" in conv.content or "france" in conv.content.lower()



"""Tests for filesystem backend fixes.

Covers:
- F1: Write serialization via threading.Lock
- F2: Append-only log writes (O(N) instead of O(N^2))
"""

import asyncio
import json
import tempfile
from datetime import UTC, datetime
from pathlib import Path
from uuid import uuid4


from ai_pipeline_core.database._types import LogRecord
from ai_pipeline_core.database.filesystem._backend import FilesystemDatabase

_DEPLOY_ID = uuid4()
_SPAN_ID = uuid4()


def _make_fs_db(tmp_path: Path) -> FilesystemDatabase:
    return FilesystemDatabase(tmp_path / "test-deploy")


def _make_log(i: int) -> LogRecord:
    return LogRecord(
        deployment_id=_DEPLOY_ID,
        span_id=_SPAN_ID,
        timestamp=datetime.now(UTC),
        sequence_no=i,
        level="INFO",
        category="test",
        logger_name=f"test.module.{i}",
        message=f"Log message {i}",
    )


# ── F2: Append-only log writes ──────────────────────────────────────────────


async def test_log_file_append_only_io() -> None:
    """_save_logs_batch_sync appends only new batch, not full accumulated logs."""
    from unittest.mock import patch

    with tempfile.TemporaryDirectory() as tmpdir:
        db = _make_fs_db(Path(tmpdir))

        total_bytes_written = 0

        original_open = Path.open

        def tracking_open(self_path: Path, *args: object, **kwargs: object):  # type: ignore[no-untyped-def]
            handle = original_open(self_path, *args, **kwargs)
            if str(self_path).endswith("logs.jsonl") and args and args[0] == "a":

                class TrackingWriter:
                    def __init__(self, h):  # type: ignore[no-untyped-def]
                        self._h = h

                    def write(self, data: str) -> int:
                        nonlocal total_bytes_written
                        total_bytes_written += len(data.encode("utf-8"))
                        return self._h.write(data)

                    def __enter__(self):  # type: ignore[no-untyped-def]
                        self._h.__enter__()
                        return self

                    def __exit__(self, *exc):  # type: ignore[no-untyped-def]
                        return self._h.__exit__(*exc)

                return TrackingWriter(handle)
            return handle

        batch_size = 10
        num_batches = 10

        with patch.object(Path, "open", tracking_open):
            for batch in range(num_batches):
                logs = [_make_log(batch * batch_size + i) for i in range(batch_size)]
                await db.save_logs_batch(logs)

        final_size = db._logs_path.stat().st_size
        ratio = total_bytes_written / final_size
        assert ratio < 2.0, (
            f"Total bytes written ({total_bytes_written}) is {ratio:.1f}x the final file size ({final_size}). Expected ratio < 2.0 for append-only writes."
        )


async def test_log_file_content_correct_after_multiple_batches() -> None:
    """All log messages must be present after multiple append batches."""
    with tempfile.TemporaryDirectory() as tmpdir:
        db = _make_fs_db(Path(tmpdir))

        batch1 = [_make_log(i) for i in range(3)]
        batch2 = [_make_log(i + 3) for i in range(3)]
        await db.save_logs_batch(batch1)
        await db.save_logs_batch(batch2)

        content = db._logs_path.read_text().strip()
        lines = content.split("\n")
        messages = {json.loads(line)["message"] for line in lines if line.strip()}
        for i in range(6):
            assert f"Log message {i}" in messages


# ── F1: Write serialization ─────────────────────────────────────────────────


async def test_concurrent_log_writes_no_data_loss() -> None:
    """Parallel save_logs_batch calls must not lose data (write lock prevents corruption)."""
    with tempfile.TemporaryDirectory() as tmpdir:
        db = _make_fs_db(Path(tmpdir))
        num_tasks = 20

        batch_size = 5

        async def write_batch(task_id: int) -> None:
            logs = [_make_log(task_id * batch_size + i) for i in range(batch_size)]
            await db.save_logs_batch(logs)

        await asyncio.gather(*(write_batch(i) for i in range(num_tasks)))

        content = db._logs_path.read_text().strip()
        lines = [line for line in content.split("\n") if line.strip()]
        messages = {json.loads(line)["message"] for line in lines}
        expected = {f"Log message {i}" for i in range(num_tasks * batch_size)}
        missing = expected - messages
        assert not missing, f"Lost {len(missing)} log messages under concurrent writes: {sorted(missing)[:5]}..."



"""Tests for resume/cached flow behavior with real Pub/Sub emulator.

Verifies that resumed (cached) flows publish correct flow.skipped events
with proper step numbering.
"""

# pyright: reportPrivateUsage=false, reportArgumentType=false

import pytest
from google.api_core.exceptions import GoogleAPICallError

from ai_pipeline_core.database._memory import _MemoryDatabase
from ai_pipeline_core.deployment._pubsub import PubSubPublisher
from ai_pipeline_core.deployment._types import EventType, _NoopPublisher

from .conftest import (
    CollectedEvent,
    PubsubTestResources,
    PublisherWithStore,
    SingleStageDeployment,
    ThreeStageDeployment,
    TwoStageDeployment,
    _flow_executions,
    make_input_doc,
    pull_events,
    run_pipeline,
)

pytestmark = pytest.mark.pubsub


def _skipped_events(events: list[CollectedEvent]) -> list[CollectedEvent]:
    """Filter to only flow.skipped events."""
    return [e for e in events if e.event_type == EventType.FLOW_SKIPPED]


def _make_fresh_publisher(pubsub_test_resources: PubsubTestResources) -> PublisherWithStore:
    """Create a new PubSubPublisher for the second run."""
    topic_id = pubsub_test_resources.topic_path.split("/")[-1]
    publisher = PubSubPublisher(
        project_id=pubsub_test_resources.project_id,
        topic_id=topic_id,
        service_type="test-service",
    )
    return PublisherWithStore(publisher=publisher)


class TestResumedFlowCachedStatus:
    """Verify that resumed flows publish flow.skipped events."""

    async def test_resumed_flow_publishes_cached_status(
        self,
        pubsub_emulator: str,
    ):
        """Run TwoStageDeployment twice; second run's skipped events have correct step numbering."""
        deployment = TwoStageDeployment()
        doc = make_input_doc()
        db = _MemoryDatabase()

        # First run: use _NoopPublisher (we don't care about its events)
        await run_pipeline(deployment, _NoopPublisher(), docs=[doc], database=db)
        assert len(_flow_executions) == 2  # both flows executed

        # Create a fresh topic + subscription for the second run
        from uuid import uuid4

        from google.cloud.pubsub_v1 import PublisherClient, SubscriberClient

        pub_client = PublisherClient()
        sub_client = SubscriberClient()
        topic_id = f"test-resume-{uuid4().hex[:8]}"
        topic_path = pub_client.topic_path("test-project", topic_id)
        sub_id = f"test-resume-sub-{uuid4().hex[:8]}"
        sub_path = sub_client.subscription_path("test-project", sub_id)
        pub_client.create_topic(name=topic_path)
        sub_client.create_subscription(name=sub_path, topic=topic_path)

        resources = PubsubTestResources(
            project_id="test-project",
            topic_path=topic_path,
            subscription_path=sub_path,
            publisher_client=pub_client,
            subscriber_client=sub_client,
        )

        try:
            second_pub = _make_fresh_publisher(resources)

            # Second run: same run_id, same docs, same database (flows should be cached)
            _flow_executions.clear()
            await run_pipeline(deployment, second_pub.publisher, docs=[doc], database=db)
            assert len(_flow_executions) == 0  # no flows executed (all cached)

            # 2 cached flows: 1 run.started + 2*flow.skipped + 1 run.completed = 4
            events = pull_events(resources, expected_count=4)
            skipped = _skipped_events(events)

            assert len(skipped) == 2

            # Step numbering is correct
            steps = sorted(evt.data["step"] for evt in skipped)
            assert steps == [1, 2]
        finally:
            try:
                sub_client.delete_subscription(subscription=sub_path)
            except OSError, GoogleAPICallError:
                pass
            try:
                pub_client.delete_topic(topic=topic_path)
            except OSError, GoogleAPICallError:
                pass


class TestResumedPipelineLifecycle:
    """Verify that fully cached pipelines still publish started + completed."""

    async def test_resumed_pipeline_still_publishes_started_and_completed(
        self,
        pubsub_emulator: str,
    ):
        """SingleStageDeployment run twice: second run still has run.started and run.completed."""
        deployment = SingleStageDeployment()
        doc = make_input_doc()
        db = _MemoryDatabase()

        # First run: use _NoopPublisher
        await run_pipeline(deployment, _NoopPublisher(), docs=[doc], database=db)
        assert len(_flow_executions) == 1

        # Create fresh topic for second run
        from uuid import uuid4

        from google.cloud.pubsub_v1 import PublisherClient, SubscriberClient

        pub_client = PublisherClient()
        sub_client = SubscriberClient()
        topic_id = f"test-resume-single-{uuid4().hex[:8]}"
        topic_path = pub_client.topic_path("test-project", topic_id)
        sub_id = f"test-resume-single-sub-{uuid4().hex[:8]}"
        sub_path = sub_client.subscription_path("test-project", sub_id)
        pub_client.create_topic(name=topic_path)
        sub_client.create_subscription(name=sub_path, topic=topic_path)

        resources = PubsubTestResources(
            project_id="test-project",
            topic_path=topic_path,
            subscription_path=sub_path,
            publisher_client=pub_client,
            subscriber_client=sub_client,
        )

        try:
            second_pub = _make_fresh_publisher(resources)

            _flow_executions.clear()
            await run_pipeline(deployment, second_pub.publisher, docs=[doc], database=db)
            assert len(_flow_executions) == 0  # cached

            # 1 cached flow: 1 run.started + flow.skipped + 1 run.completed = 3
            events = pull_events(resources, expected_count=3)

            started = [e for e in events if e.event_type == EventType.RUN_STARTED]
            completed = [e for e in events if e.event_type == EventType.RUN_COMPLETED]
            skipped = _skipped_events(events)

            assert len(started) == 1
            assert len(completed) == 1
            assert len(skipped) == 1
        finally:
            try:
                sub_client.delete_subscription(subscription=sub_path)
            except OSError, GoogleAPICallError:
                pass
            try:
                pub_client.delete_topic(topic=topic_path)
            except OSError, GoogleAPICallError:
                pass


class TestPartialResumeMix:
    """Verify fully cached multi-flow pipelines publish correct skipped events."""

    async def test_partial_resume_mix_of_cached_and_executed(
        self,
        pubsub_emulator: str,
    ):
        """3-flow deployment: all flows have completion records from first run,
        all get flow.skipped on second run.
        """
        deployment = ThreeStageDeployment()
        doc = make_input_doc()
        db = _MemoryDatabase()

        # First run to populate the database with documents and completions
        await run_pipeline(deployment, _NoopPublisher(), docs=[doc], database=db)
        assert set(_flow_executions) == {"flow_a", "flow_b", "flow_c"}

        # Create fresh topic for second run
        from uuid import uuid4

        from google.cloud.pubsub_v1 import PublisherClient, SubscriberClient

        pub_client = PublisherClient()
        sub_client = SubscriberClient()
        topic_id = f"test-partial-resume-{uuid4().hex[:8]}"
        topic_path = pub_client.topic_path("test-project", topic_id)
        sub_id = f"test-partial-resume-sub-{uuid4().hex[:8]}"
        sub_path = sub_client.subscription_path("test-project", sub_id)
        pub_client.create_topic(name=topic_path)
        sub_client.create_subscription(name=sub_path, topic=topic_path)

        resources = PubsubTestResources(
            project_id="test-project",
            topic_path=topic_path,
            subscription_path=sub_path,
            publisher_client=pub_client,
            subscriber_client=sub_client,
        )

        try:
            second_pub = _make_fresh_publisher(resources)

            _flow_executions.clear()
            await run_pipeline(deployment, second_pub.publisher, docs=[doc], database=db)

            # All flows should be cached (no re-execution)
            assert _flow_executions == []

            # 3 cached flows: 1 run.started + 3*flow.skipped + 1 run.completed = 5
            events = pull_events(resources, expected_count=5)

            skipped = _skipped_events(events)
            assert len(skipped) == 3

            # All three flows: skipped
            skipped_names = sorted(evt.data["flow_name"] for evt in skipped)
            assert "chain_input_to_middle" in skipped_names
            assert "chain_middle_to_output" in skipped_names
            assert "chain_output_to_final" in skipped_names

            # No flow.started or flow.completed events
            flow_started = [e for e in events if e.event_type == EventType.FLOW_STARTED]
            flow_completed = [e for e in events if e.event_type == EventType.FLOW_COMPLETED]
            assert len(flow_started) == 0
            assert len(flow_completed) == 0
        finally:
            try:
                sub_client.delete_subscription(subscription=sub_path)
            except OSError, GoogleAPICallError:
                pass
            try:
                pub_client.delete_topic(topic=topic_path)
            except OSError, GoogleAPICallError:
                pass



"""AI Pipeline Core - Production-ready framework for building AI pipelines with LLMs."""

import importlib.metadata
import os
import sys

# Disable Prefect telemetry and analytics before importing Prefect.
# Execution tracking is handled by the framework database.
os.environ.setdefault("DO_NOT_TRACK", "1")
os.environ.setdefault("PREFECT_CLOUD_ENABLE_ORCHESTRATION_TELEMETRY", "false")

from prefect.context import refresh_global_settings_context
from prefect.settings import get_current_settings

# If Prefect was already imported (user imported it before us), refresh its cached settings.
if "prefect" in sys.modules and get_current_settings().cloud.enable_orchestration_telemetry:
    refresh_global_settings_context()

from . import llm
from ._codec import CodecError, UniversalCodec
from .database import DatabaseReader
from .deployment import DeploymentResult, PipelineDeployment
from .deployment.remote import RemoteDeployment
from .documents import (
    Attachment,
    Document,
    DocumentSha256,
    ensure_extension,
    find_all,
    find_document,
    find_latest,
    is_document_sha256,
    replace_extension,
    sanitize_url,
)
from .exceptions import (
    DocumentNameError,
    DocumentSizeError,
    DocumentValidationError,
    LLMError,
    OutputDegenerationError,
    PipelineCoreError,
)
from .llm import (
    Citation,
    Conversation,
    ModelName,
    ModelOptions,
    TokenUsage,
    Tool,
    ToolOutput,
)
from .logger import (
    LoggingConfig,
    get_pipeline_logger,
    setup_logging,
)
from .pipeline import (
    FlowOptions,
    LimitKind,
    PipelineFlow,
    PipelineLimit,
    PipelineTask,
    TaskBatch,
    TaskHandle,
    add_cost,
    as_task_completed,
    collect_tasks,
    get_run_id,
    pipeline_concurrency,
    pipeline_test_context,
    run_tasks_until,
    safe_gather,
    safe_gather_indexed,
    traced_operation,
)
from .prompt_compiler import Guide, ListField, MultiLineField, OutputRule, OutputT, PromptSpec, Role, Rule, StructuredField, render_preview, render_text
from .replay import ExperimentOverrides, ExperimentResult, execute_span, experiment_batch, experiment_span
from .settings import Settings
from .testing import disable_run_logger, prefect_test_harness

__version__ = importlib.metadata.version("ai-pipeline-core")

__all__ = [
    "Attachment",
    "Citation",
    "CodecError",
    "Conversation",
    "DatabaseReader",
    "DeploymentResult",
    "Document",
    "DocumentNameError",
    "DocumentSha256",
    "DocumentSizeError",
    "DocumentValidationError",
    "ExperimentOverrides",
    "ExperimentResult",
    "FlowOptions",
    "Guide",
    "LLMError",
    "LimitKind",
    "ListField",
    "LoggingConfig",
    "ModelName",
    "ModelOptions",
    "MultiLineField",
    "OutputDegenerationError",
    "OutputRule",
    "OutputT",
    "PipelineCoreError",
    "PipelineDeployment",
    "PipelineFlow",
    "PipelineLimit",
    "PipelineTask",
    "PromptSpec",
    "RemoteDeployment",
    "Role",
    "Rule",
    "Settings",
    "StructuredField",
    "TaskBatch",
    "TaskHandle",
    "TokenUsage",
    "Tool",
    "ToolOutput",
    "UniversalCodec",
    "add_cost",
    "as_task_completed",
    "collect_tasks",
    "disable_run_logger",
    "ensure_extension",
    "execute_span",
    "experiment_batch",
    "experiment_span",
    "find_all",
    "find_document",
    "find_latest",
    "get_pipeline_logger",
    "get_run_id",
    "is_document_sha256",
    "llm",
    "pipeline_concurrency",
    "pipeline_test_context",
    "prefect_test_harness",
    "render_preview",
    "render_text",
    "replace_extension",
    "run_tasks_until",
    "safe_gather",
    "safe_gather_indexed",
    "sanitize_url",
    "setup_logging",
    "traced_operation",
]



# pyright: reportPrivateUsage=false
"""Tests for plan_next_flow skip/continue logic and resolve_document_inputs provenance rejection."""

from collections.abc import Sequence

import pytest

from ai_pipeline_core import DeploymentResult, Document, FlowOptions, PipelineDeployment
from ai_pipeline_core.deployment import FlowAction, FlowDirective
from ai_pipeline_core.deployment._types import _MemoryPublisher
from ai_pipeline_core.deployment._resolve import _DocumentInput
from ai_pipeline_core.deployment._resolve import resolve_document_inputs
from ai_pipeline_core.deployment._types import FlowSkippedEvent
from ai_pipeline_core.pipeline import PipelineFlow, PipelineTask


# ---------------------------------------------------------------------------
# Document types
# ---------------------------------------------------------------------------


class PlanInputDoc(Document):
    """Input for plan_next_flow tests."""


class PlanMiddleDoc(Document):
    """Intermediate output for plan_next_flow tests."""


class PlanOutputDoc(Document):
    """Final output for plan_next_flow tests."""


class PlanResult(DeploymentResult):
    output_count: int = 0


# ---------------------------------------------------------------------------
# Tasks + Flows
# ---------------------------------------------------------------------------


class ToMiddleTask(PipelineTask):
    @classmethod
    async def run(cls, documents: tuple[PlanInputDoc, ...]) -> tuple[PlanMiddleDoc, ...]:
        return tuple(PlanMiddleDoc.derive(derived_from=(d,), name=f"mid_{d.name}", content="mid") for d in documents)


class ToOutputTask(PipelineTask):
    @classmethod
    async def run(cls, documents: tuple[PlanMiddleDoc, ...]) -> tuple[PlanOutputDoc, ...]:
        return tuple(PlanOutputDoc.derive(derived_from=(d,), name=f"out_{d.name}", content="out") for d in documents)


class ProducerFlow(PipelineFlow):
    name = "producer"

    async def run(self, documents: tuple[PlanInputDoc, ...], options: FlowOptions) -> tuple[PlanMiddleDoc, ...]:
        return await ToMiddleTask.run(documents)


class EmptyProducerFlow(PipelineFlow):
    name = "empty-producer"

    async def run(self, documents: tuple[PlanInputDoc, ...], options: FlowOptions) -> tuple[PlanMiddleDoc, ...]:
        return ()


class ConsumerFlow(PipelineFlow):
    name = "consumer"
    ran = False

    async def run(self, documents: tuple[PlanMiddleDoc, ...], options: FlowOptions) -> tuple[PlanOutputDoc, ...]:
        type(self).ran = True
        return await ToOutputTask.run(documents)


# ---------------------------------------------------------------------------
# plan_next_flow with actual previous_output_documents
# ---------------------------------------------------------------------------


class _SkipWhenEmptyDeployment(PipelineDeployment[FlowOptions, PlanResult]):
    """Skips consumer flow when producer outputs zero documents."""

    def build_flows(self, options: FlowOptions) -> list[PipelineFlow]:
        return [EmptyProducerFlow(), ConsumerFlow()]

    def plan_next_flow(
        self,
        flow_class: type[PipelineFlow],
        plan: Sequence[PipelineFlow],
        output_documents: tuple[Document, ...],
    ) -> FlowDirective:
        if flow_class is ConsumerFlow and not output_documents:
            return FlowDirective(action=FlowAction.SKIP, reason="no intermediate documents")
        return FlowDirective()

    @staticmethod
    def build_result(run_id: str, documents: tuple[Document, ...], options: FlowOptions) -> PlanResult:
        return PlanResult(success=True, output_count=len(documents))


class _ContinueWhenDocsExist(PipelineDeployment[FlowOptions, PlanResult]):
    """Continues consumer flow when producer produces documents."""

    def build_flows(self, options: FlowOptions) -> list[PipelineFlow]:
        return [ProducerFlow(), ConsumerFlow()]

    def plan_next_flow(
        self,
        flow_class: type[PipelineFlow],
        plan: Sequence[PipelineFlow],
        output_documents: tuple[Document, ...],
    ) -> FlowDirective:
        if flow_class is ConsumerFlow and not output_documents:
            return FlowDirective(action=FlowAction.SKIP, reason="no intermediate documents")
        return FlowDirective()

    @staticmethod
    def build_result(run_id: str, documents: tuple[Document, ...], options: FlowOptions) -> PlanResult:
        return PlanResult(success=True, output_count=len(documents))


@pytest.mark.asyncio
async def test_plan_next_flow_skips_when_no_output_docs() -> None:
    """plan_next_flow receives empty output_documents and skips next flow."""
    publisher = _MemoryPublisher()
    ConsumerFlow.ran = False
    doc = PlanInputDoc.create_root(name="in.txt", content="x", reason="gap7")
    result = await _SkipWhenEmptyDeployment().run("gap7-skip", [doc], FlowOptions(), publisher=publisher)

    assert result.success
    assert not ConsumerFlow.ran

    skipped = [e for e in publisher.events if isinstance(e, FlowSkippedEvent)]
    assert len(skipped) == 1
    assert skipped[0].flow_name == "consumer"
    assert "no intermediate documents" in skipped[0].reason


@pytest.mark.asyncio
async def test_plan_next_flow_continues_when_output_docs_exist() -> None:
    """plan_next_flow receives output_documents and continues (doesn't skip)."""
    publisher = _MemoryPublisher()
    ConsumerFlow.ran = False
    doc = PlanInputDoc.create_root(name="in.txt", content="x", reason="gap7b")
    result = await _ContinueWhenDocsExist().run("gap7-continue", [doc], FlowOptions(), publisher=publisher)

    assert result.success
    assert ConsumerFlow.ran

    skipped = [e for e in publisher.events if isinstance(e, FlowSkippedEvent)]
    assert skipped == []


# ---------------------------------------------------------------------------
# resolve_document_inputs provenance rejection
# ---------------------------------------------------------------------------


class ResolveInputDoc(Document):
    """Document type for resolve test."""


@pytest.mark.asyncio
async def test_resolve_rejects_derived_from_on_input() -> None:
    """DocumentInput with derived_from raises ValueError."""
    inputs = [_DocumentInput(content="hello", name="x.txt", class_name="ResolveInputDoc", derived_from=("SOMESHA256",))]
    with pytest.raises(ValueError, match="cannot set derived_from"):
        await resolve_document_inputs(inputs, [ResolveInputDoc])


@pytest.mark.asyncio
async def test_resolve_rejects_triggered_by_on_input() -> None:
    """DocumentInput with triggered_by raises ValueError."""
    inputs = [_DocumentInput(content="hello", name="x.txt", class_name="ResolveInputDoc", triggered_by=("SOMESHA256",))]
    with pytest.raises(ValueError, match="cannot set derived_from"):
        await resolve_document_inputs(inputs, [ResolveInputDoc])



"""Shared ClickHouse async client helpers."""

import asyncio
import warnings
from dataclasses import dataclass
from importlib.metadata import PackageNotFoundError
from importlib.metadata import version as package_version

from clickhouse_connect import get_async_client
from clickhouse_connect.driver.asyncclient import AsyncClient
from clickhouse_connect.driver.exceptions import DatabaseError as ClickHouseDatabaseError

from ai_pipeline_core.database.clickhouse._ddl import DDL_STATEMENTS, SCHEMA_META_TABLE, SCHEMA_VERSION
from ai_pipeline_core.settings import Settings

# clickhouse-connect 0.14+ warns that the thread-pool async wrapper will be replaced
# by a native async client in 1.0. Suppress globally until we migrate.
warnings.filterwarnings("ignore", category=FutureWarning, module=r"clickhouse_connect")

__all__ = [
    "SchemaVersionError",
    "get_async_clickhouse_client",
    "reset_schema_check",
]


@dataclass(slots=True)
class _SchemaState:
    """Process-level schema verification flag."""

    verified: bool = False


_schema_state = _SchemaState()
_schema_lock = asyncio.Lock()


async def _table_exists(client: AsyncClient, table: str, database: str) -> bool:
    result = await client.query(
        "SELECT 1 FROM system.tables WHERE database = {db:String} AND name = {tbl:String}",
        parameters={"db": database, "tbl": table},
    )
    return len(result.result_rows) > 0


async def _ensure_schema(client: AsyncClient, database: str) -> None:
    """Check schema version or auto-initialize an empty database.

    - No schema_meta table → fresh database → create all tables and stamp version.
    - schema_meta exists, version matches → proceed.
    - schema_meta exists, DB version < framework → raise error (DB needs upgrade).
    - schema_meta exists, DB version > framework → raise error (framework needs upgrade).
    """
    if _schema_state.verified:
        return

    async with _schema_lock:
        if _schema_state.verified:
            return

        meta_exists = await _table_exists(client, SCHEMA_META_TABLE, database)

        if not meta_exists:
            fw_version = _get_framework_version()
            for ddl in DDL_STATEMENTS:
                await client.command(ddl)
            await client.command(
                f"INSERT INTO {SCHEMA_META_TABLE} (version, applied_at, framework_version) VALUES ({{version:UInt32}}, now64(3), {{fw:String}})",
                parameters={"version": SCHEMA_VERSION, "fw": fw_version},
            )
            _schema_state.verified = True
            return

        result = await client.query(f"SELECT max(version) FROM {SCHEMA_META_TABLE}")
        db_version = int(result.result_rows[0][0]) if result.result_rows and result.result_rows[0][0] else 0

        if db_version < SCHEMA_VERSION:
            raise SchemaVersionError(
                f"Database schema version ({db_version}) is older than the framework expects ({SCHEMA_VERSION}). "
                "Update the database schema to match the current framework version."
            )
        if db_version > SCHEMA_VERSION:
            raise SchemaVersionError(
                f"Database schema version ({db_version}) is newer than the framework supports ({SCHEMA_VERSION}). "
                f"Upgrade ai-pipeline-core to a version that supports schema version {db_version}."
            )

        _schema_state.verified = True


def _get_framework_version() -> str:
    try:
        return package_version("ai-pipeline-core")
    except PackageNotFoundError:
        return "unknown"


class SchemaVersionError(Exception):
    """Raised when the database schema version does not match the framework's expected version."""


def reset_schema_check() -> None:
    """Reset the process-level schema verification flag. For use in tests only."""
    _schema_state.verified = False


async def get_async_clickhouse_client(settings: Settings | None = None) -> AsyncClient:
    """Create a new ClickHouse async client and verify schema version."""
    active_settings = settings or Settings()
    client = await get_async_client(
        host=active_settings.clickhouse_host,
        port=active_settings.clickhouse_port,
        database=active_settings.clickhouse_database,
        username=active_settings.clickhouse_user,
        password=active_settings.clickhouse_password,
        secure=active_settings.clickhouse_secure,
        connect_timeout=active_settings.clickhouse_connect_timeout,
        send_receive_timeout=active_settings.clickhouse_send_receive_timeout,
    )
    try:
        await _ensure_schema(client, active_settings.clickhouse_database)
    except ClickHouseDatabaseError:
        await client.close()
        raise
    return client



"""CLI usage tests for generic replay."""

from pathlib import Path
from types import SimpleNamespace
from typing import Any
from uuid import uuid4

import pytest

from ai_pipeline_core.database.filesystem._backend import FilesystemDatabase
from ai_pipeline_core.replay.cli import infer_db_path, main
from tests.replay.conftest import make_span


class _MockResult:
    def __init__(self, content: str = "LLM response") -> None:
        self.content = content
        self.usage = SimpleNamespace(total_tokens=150, model_dump=lambda: {"total_tokens": 150})
        self.cost = 0.003


async def _seed_span(database: FilesystemDatabase, span: object) -> None:
    await database.insert_span(span)


def _conversation_span(tmp_path: Path) -> tuple[FilesystemDatabase, object]:
    database = FilesystemDatabase(tmp_path / "bundle")
    span = make_span(
        kind="conversation",
        name="ReplayConversation",
        target="decoded_method:ai_pipeline_core.llm.conversation:Conversation.send",
        meta={"model": "gemini-3-flash", "purpose": "summary"},
        metrics={"time_taken_ms": 123},
    )
    import asyncio

    asyncio.run(_seed_span(database, span))
    return database, span


@pytest.mark.ai_docs
def test_main_show_from_db_displays_meta_and_metrics(tmp_path: Path, capsys: pytest.CaptureFixture[str]) -> None:
    _database, span = _conversation_span(tmp_path)

    exit_code = main(["show", "--from-db", str(span.span_id), "--db-path", str(tmp_path / "bundle")])

    assert exit_code == 0
    output = capsys.readouterr().out
    assert "meta_json" in output
    assert "metrics_json" in output
    assert "gemini-3-flash" in output


@pytest.mark.ai_docs
def test_main_run_from_file_writes_output(tmp_path: Path, monkeypatch: pytest.MonkeyPatch) -> None:
    _database, span = _conversation_span(tmp_path)
    replay_file = tmp_path / "bundle" / "spans" / f"{span.span_id}.json"

    async def fake_execute(span_id: Any, *, source_db: Any, sink_db: Any = None) -> _MockResult:
        _ = (span_id, source_db, sink_db)
        return _MockResult()

    monkeypatch.setattr("ai_pipeline_core.replay.cli.execute_span", fake_execute)

    exit_code = main(["run", str(replay_file), "--db-path", str(tmp_path / "bundle")])

    assert exit_code == 0
    output_dir = replay_file.parent / f"{replay_file.stem}_replay"
    assert (output_dir / "output.yaml").exists()


@pytest.mark.ai_docs
def test_main_run_with_overrides_uses_experiment_span(tmp_path: Path, monkeypatch: pytest.MonkeyPatch) -> None:
    _database, span = _conversation_span(tmp_path)
    replay_file = tmp_path / "bundle" / "spans" / f"{span.span_id}.json"

    async def fail_execute(span_id: Any, *, source_db: Any, sink_db: Any = None) -> _MockResult:
        _ = (span_id, source_db, sink_db)
        raise AssertionError("run with overrides should use experiment_span")

    async def fake_experiment_span(
        span_id: Any,
        *,
        source_db: Any,
        sink_db: Any = None,
        overrides: Any = None,
    ) -> Any:
        _ = (source_db, sink_db)
        assert span_id == span.span_id
        assert overrides is not None
        assert overrides.model == "gemini-3-flash"
        assert overrides.model_options == {"reasoning_effort": "low"}
        return SimpleNamespace(result=_MockResult("override response"))

    monkeypatch.setattr("ai_pipeline_core.replay.cli.execute_span", fail_execute)
    monkeypatch.setattr("ai_pipeline_core.replay.cli.experiment_span", fake_experiment_span)

    exit_code = main([
        "run",
        str(replay_file),
        "--db-path",
        str(tmp_path / "bundle"),
        "--model",
        "gemini-3-flash",
        "--set",
        "reasoning_effort=low",
    ])

    assert exit_code == 0
    output_dir = replay_file.parent / f"{replay_file.stem}_replay"
    assert (output_dir / "output.yaml").exists()


@pytest.mark.ai_docs
def test_main_batch_uses_find_and_experiment_helpers(tmp_path: Path, monkeypatch: pytest.MonkeyPatch, capsys: pytest.CaptureFixture[str]) -> None:
    _database, span = _conversation_span(tmp_path)
    deployment_id = span.root_deployment_id

    async def fake_find(
        database: Any,
        deployment_id_arg: Any,
        *,
        kind: Any = None,
        purpose: Any = None,
        task_class: Any = None,
    ) -> list[Any]:
        _ = database
        assert deployment_id_arg == deployment_id
        assert kind == "conversation"
        assert purpose == "summary"
        assert task_class is None
        return [span.span_id]

    async def fake_batch(
        span_ids: Any,
        source_db: Any,
        *,
        overrides: Any = None,
        concurrency: int = 5,
        sink_db: Any = None,
    ) -> list[Any]:
        _ = (source_db, sink_db)
        assert span_ids == [span.span_id]
        assert overrides is not None
        assert overrides.model == "gemini-3-flash"
        assert overrides.model_options == {"reasoning_effort": "low"}
        assert concurrency == 3
        return [
            SimpleNamespace(
                source_span_id=span.span_id,
                model_used="gemini-3-flash",
                duration_seconds=0.5,
                cost_usd=0.0,
            )
        ]

    monkeypatch.setattr("ai_pipeline_core.replay.cli.find_experiment_span_ids", fake_find)
    monkeypatch.setattr("ai_pipeline_core.replay.cli.experiment_batch", fake_batch)

    exit_code = main([
        "batch",
        "--from-deployment",
        str(deployment_id),
        "--db-path",
        str(tmp_path / "bundle"),
        "--kind",
        "conversation",
        "--purpose",
        "summary",
        "--model",
        "gemini-3-flash",
        "--set",
        "reasoning_effort=low",
        "--concurrency",
        "3",
    ])

    assert exit_code == 0
    assert "Ran 1 replay experiments" in capsys.readouterr().out


def test_infer_db_path_finds_span_snapshot_root(tmp_path: Path) -> None:
    snapshot_root = tmp_path / "snapshot"
    (snapshot_root / "spans").mkdir(parents=True)
    replay_file = snapshot_root / "spans" / f"{uuid4()}.json"
    replay_file.write_text("{}", encoding="utf-8")

    assert infer_db_path(replay_file) == snapshot_root



"""Tests for LaminarSpanSink."""

from pathlib import Path
from types import MappingProxyType
from uuid import uuid7

import pytest

from ai_pipeline_core.database import SpanKind
from ai_pipeline_core.deployment._types import _NoopPublisher
from ai_pipeline_core.observability._laminar_sink import LaminarSpanSink, _reset_for_testing
from ai_pipeline_core.pipeline._execution_context import ExecutionContext
from ai_pipeline_core.pipeline._runtime_sinks import build_runtime_sinks
from ai_pipeline_core.pipeline._span_sink import SpanMetrics
from ai_pipeline_core.pipeline.limits import _SharedStatus
from ai_pipeline_core.settings import Settings


class _FakeSpan:
    def __init__(self, record: dict[str, object]) -> None:
        self._record = record

    def get_laminar_span_context(self) -> str:
        if _FakeLaminar.fail_get_context:
            raise RuntimeError("get context failed")
        return self._record["context"]  # type: ignore[return-value]

    def set_attributes(self, attributes: dict[str, object]) -> None:
        if _FakeLaminar.fail_set_attributes:
            raise RuntimeError("set_attributes failed")
        self._record["attribute_updates"] = attributes

    def set_output(self, output: object) -> None:
        if _FakeLaminar.fail_set_output:
            raise RuntimeError("set_output failed")
        self._record["output"] = output

    def end(self) -> None:
        if _FakeLaminar.fail_end:
            raise RuntimeError("end failed")
        self._record["ended"] = True


class _FakeLaminar:
    initialize_calls: list[dict[str, object]] = []
    start_calls: list[dict[str, object]] = []
    fail_initialize = False
    fail_start = False
    fail_get_context = False
    fail_set_attributes = False
    fail_set_output = False
    fail_end = False

    @classmethod
    def initialize(cls, *, project_api_key: str, disabled_instruments: set[object]) -> None:
        if cls.fail_initialize:
            raise RuntimeError("initialize failed")
        cls.initialize_calls.append({
            "project_api_key": project_api_key,
            "disabled_instruments": disabled_instruments,
        })

    @classmethod
    def start_span(
        cls,
        *,
        name: str,
        span_type: str,
        parent_span_context: object,
        input: object,
        attributes: dict[str, object],
    ) -> _FakeSpan:
        if cls.fail_start:
            raise RuntimeError("start failed")
        record: dict[str, object] = {
            "name": name,
            "span_type": span_type,
            "parent_span_context": parent_span_context,
            "input": input,
            "attributes": attributes,
            "context": f"context-{len(cls.start_calls)}",
            "ended": False,
        }
        cls.start_calls.append(record)
        return _FakeSpan(record)


class _FakeInstruments:
    OPENAI = "openai"
    LITELLM = "litellm"


def _install_fake_lmnr(monkeypatch: pytest.MonkeyPatch) -> None:
    from ai_pipeline_core.observability import _laminar_sink

    monkeypatch.setattr(_laminar_sink, "Laminar", _FakeLaminar)
    monkeypatch.setattr(_laminar_sink, "Instruments", _FakeInstruments)


@pytest.fixture(autouse=True)
def _reset_state() -> None:
    _reset_for_testing()
    _FakeLaminar.initialize_calls.clear()
    _FakeLaminar.start_calls.clear()
    _FakeLaminar.fail_initialize = False
    _FakeLaminar.fail_start = False
    _FakeLaminar.fail_get_context = False
    _FakeLaminar.fail_set_attributes = False
    _FakeLaminar.fail_set_output = False
    _FakeLaminar.fail_end = False


def _build_context() -> ExecutionContext:
    return ExecutionContext(
        run_id="test-run",
        execution_id=None,
        publisher=_NoopPublisher(),
        limits=MappingProxyType({}),
        limits_status=_SharedStatus(),
    )


def test_build_runtime_sinks_includes_laminar_sink_when_key_is_set() -> None:
    sinks = build_runtime_sinks(database=None, settings_obj=Settings(lmnr_project_api_key="secret"))

    assert len(sinks) == 1
    assert isinstance(sinks[0], LaminarSpanSink)


@pytest.mark.asyncio
async def test_laminar_sink_maps_llm_round_to_gen_ai_attributes(monkeypatch: pytest.MonkeyPatch) -> None:
    _install_fake_lmnr(monkeypatch)
    sink = LaminarSpanSink(Settings(lmnr_project_api_key="secret"))
    parent_span_id = uuid7()
    child_span_id = uuid7()

    await sink.on_span_started(
        span_id=parent_span_id,
        parent_span_id=None,
        kind=SpanKind.CONVERSATION,
        name="conversation",
        target="decoded_method:ai_pipeline_core.llm.conversation:Conversation.send",
        started_at=None,
        input_preview={"conversation": True},
    )
    await sink.on_span_started(
        span_id=child_span_id,
        parent_span_id=parent_span_id,
        kind=SpanKind.LLM_ROUND,
        name="round-1",
        target="",
        started_at=None,
        input_preview=[{"role": "user", "content": "hello"}],
    )
    await sink.on_span_finished(
        span_id=child_span_id,
        ended_at=None,
        output_preview="world",
        error=None,
        metrics=SpanMetrics(
            time_taken_ms=25,
            log_summary={"total": 0, "warnings": 0, "errors": 0, "last_error": ""},
            tokens_input=11,
            tokens_output=7,
            tokens_cache_read=3,
            tokens_reasoning=2,
            cost_usd=0.5,
            first_token_ms=4,
        ),
        meta={"model": "gpt-5.1", "response_id": "resp-123"},
    )

    assert _FakeLaminar.initialize_calls == [
        {
            "project_api_key": "secret",
            "disabled_instruments": {"openai", "litellm"},
        }
    ]
    assert len(_FakeLaminar.start_calls) == 2
    assert _FakeLaminar.start_calls[0]["span_type"] == "DEFAULT"
    assert _FakeLaminar.start_calls[1]["span_type"] == "LLM"
    assert _FakeLaminar.start_calls[1]["parent_span_context"] == "context-0"
    assert _FakeLaminar.start_calls[1]["attributes"] == {
        "ai_pipeline.span_kind": SpanKind.LLM_ROUND.value,
        "ai_pipeline.target": "",
    }
    assert _FakeLaminar.start_calls[1]["attribute_updates"] == {
        "gen_ai.system": "litellm",
        "gen_ai.usage.input_tokens": 11,
        "gen_ai.usage.output_tokens": 7,
        "gen_ai.usage.cache_read_input_tokens": 3,
        "gen_ai.usage.reasoning_tokens": 2,
        "gen_ai.usage.cost": 0.5,
        "gen_ai.input.messages": '[{"role": "user", "content": "hello"}]',
        "gen_ai.output.messages": '"world"',
        "gen_ai.request.model": "gpt-5.1",
        "gen_ai.response.model": "gpt-5.1",
        "gen_ai.response.id": "resp-123",
    }
    assert _FakeLaminar.start_calls[1]["output"] == "world"
    assert _FakeLaminar.start_calls[1]["ended"] is True
    assert child_span_id not in sink._open_spans


@pytest.mark.asyncio
async def test_laminar_sink_is_fail_open(monkeypatch: pytest.MonkeyPatch) -> None:
    _install_fake_lmnr(monkeypatch)
    sink = LaminarSpanSink(Settings(lmnr_project_api_key="secret"))
    span_id = uuid7()

    _FakeLaminar.fail_start = True
    await sink.on_span_started(
        span_id=span_id,
        parent_span_id=None,
        kind=SpanKind.LLM_ROUND,
        name="round-1",
        target="",
        started_at=None,
        input_preview="input",
    )
    assert sink._open_spans == {}

    _FakeLaminar.fail_start = False
    await sink.on_span_started(
        span_id=span_id,
        parent_span_id=None,
        kind=SpanKind.LLM_ROUND,
        name="round-1",
        target="",
        started_at=None,
        input_preview="input",
    )
    _FakeLaminar.fail_set_attributes = True
    _FakeLaminar.fail_set_output = True
    _FakeLaminar.fail_end = True
    await sink.on_span_finished(
        span_id=span_id,
        ended_at=None,
        output_preview="output",
        error=RuntimeError("boom"),
        metrics=SpanMetrics(
            time_taken_ms=10,
            log_summary={"total": 0, "warnings": 0, "errors": 0, "last_error": ""},
        ),
        meta={"model": "gpt-5.1", "response_id": "resp-123"},
    )


@pytest.mark.asyncio
async def test_laminar_sink_ends_span_when_context_capture_fails(monkeypatch: pytest.MonkeyPatch) -> None:
    _install_fake_lmnr(monkeypatch)
    sink = LaminarSpanSink(Settings(lmnr_project_api_key="secret"))
    span_id = uuid7()

    _FakeLaminar.fail_get_context = True
    await sink.on_span_started(
        span_id=span_id,
        parent_span_id=None,
        kind=SpanKind.LLM_ROUND,
        name="round-1",
        target="",
        started_at=None,
        input_preview="input",
    )

    assert sink._open_spans == {}
    assert _FakeLaminar.start_calls[0]["ended"] is True


@pytest.mark.asyncio
async def test_laminar_sink_skips_export_when_process_was_initialized_with_different_key(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    _install_fake_lmnr(monkeypatch)
    first_sink = LaminarSpanSink(Settings(lmnr_project_api_key="project-a"))
    second_sink = LaminarSpanSink(Settings(lmnr_project_api_key="project-b"))

    await first_sink.on_span_started(
        span_id=uuid7(),
        parent_span_id=None,
        kind=SpanKind.CONVERSATION,
        name="conversation",
        target="decoded_method:ai_pipeline_core.llm.conversation:Conversation.send",
        started_at=None,
        input_preview={"conversation": True},
    )
    await second_sink.on_span_started(
        span_id=uuid7(),
        parent_span_id=None,
        kind=SpanKind.CONVERSATION,
        name="conversation",
        target="decoded_method:ai_pipeline_core.llm.conversation:Conversation.send",
        started_at=None,
        input_preview={"conversation": True},
    )

    assert _FakeLaminar.initialize_calls == [
        {
            "project_api_key": "project-a",
            "disabled_instruments": {"openai", "litellm"},
        }
    ]
    assert len(_FakeLaminar.start_calls) == 1


def test_no_laminar_span_calls_remain() -> None:
    needle = "laminar" + "_span("
    hits: list[str] = []
    for root in (Path("ai_pipeline_core"), Path("tests")):
        for path in root.rglob("*.py"):
            if path.name == "test_laminar_sink.py":
                continue
            if needle in path.read_text(encoding="utf-8"):
                hits.append(str(path))

    assert hits == []



"""Pipeline framework primitives."""

from ai_pipeline_core.pipeline._execution_context import add_cost, get_run_id, pipeline_test_context
from ai_pipeline_core.pipeline._flow import PipelineFlow
from ai_pipeline_core.pipeline._parallel import TaskBatch, TaskHandle, as_task_completed, collect_tasks, run_tasks_until
from ai_pipeline_core.pipeline._task import PipelineTask
from ai_pipeline_core.pipeline._traced import traced_operation
from ai_pipeline_core.pipeline.gather import safe_gather, safe_gather_indexed
from ai_pipeline_core.pipeline.limits import LimitKind, PipelineLimit, pipeline_concurrency
from ai_pipeline_core.pipeline.options import FlowOptions

__all__ = [
    "FlowOptions",
    "LimitKind",
    "PipelineFlow",
    "PipelineLimit",
    "PipelineTask",
    "TaskBatch",
    "TaskHandle",
    "add_cost",
    "as_task_completed",
    "collect_tasks",
    "get_run_id",
    "pipeline_concurrency",
    "pipeline_test_context",
    "run_tasks_until",
    "safe_gather",
    "safe_gather_indexed",
    "traced_operation",
]



"""ClickHouse database backend."""

from ai_pipeline_core.database.clickhouse._backend import ClickHouseDatabase
from ai_pipeline_core.database.clickhouse._connection import SchemaVersionError

__all__ = [
    "ClickHouseDatabase",
    "SchemaVersionError",
]



"""Tests for StructuredField and ListField support in PromptSpec.

StructuredField: BaseModel values rendered as JSON (indent=2) in XML tags.
ListField: list values rendered as individually tagged items in XML wrapper.
Both are sent as user messages before the main prompt, same as MultiLineField.
"""

from unittest.mock import patch

import pytest
from pydantic import BaseModel, Field

from ai_pipeline_core._llm_core.model_response import ModelResponse
from ai_pipeline_core._llm_core.types import TokenUsage
from ai_pipeline_core.documents import Document
from ai_pipeline_core.llm import Conversation
from ai_pipeline_core.prompt_compiler.components import Role
from ai_pipeline_core.prompt_compiler.render import render_multi_line_messages, render_preview, render_text
from ai_pipeline_core.prompt_compiler.spec import (
    ListField,
    PromptSpec,
    StructuredField,
    _is_list_field,
    _is_structured_field,
)


# ---------------------------------------------------------------------------
# Test fixtures / helpers
# ---------------------------------------------------------------------------


class SLDoc(Document):
    """Test document for structured/list tests."""


class SLRole(Role):
    """Test role."""

    text = "experienced analyst"


class Finding(BaseModel):
    """A single finding."""

    model_config = {"frozen": True}

    title: str
    severity: str


class NestedDetail(BaseModel):
    """Nested detail model."""

    model_config = {"frozen": True}

    category: str
    tags: list[str]


def _make_model_response(content: str) -> ModelResponse[str]:
    return ModelResponse[str](
        content=content,
        parsed=content,
        model="test-model",
        usage=TokenUsage(prompt_tokens=10, completion_tokens=5, total_tokens=15),
    )


# ---------------------------------------------------------------------------
# StructuredField: import and declaration
# ---------------------------------------------------------------------------


def test_structured_field_importable_from_prompt_compiler() -> None:
    from ai_pipeline_core.prompt_compiler import StructuredField

    assert callable(StructuredField)


def test_structured_field_importable_from_top_level() -> None:
    from ai_pipeline_core import StructuredField

    assert callable(StructuredField)


def test_structured_field_creates_valid_spec() -> None:
    class FindingSpec(PromptSpec):
        """Analyze a finding."""

        input_documents = ()
        role = SLRole
        task = "Analyze the finding provided."

        finding: Finding = StructuredField(description="Primary finding")

    spec = FindingSpec(finding=Finding(title="Oracle attack", severity="critical"))
    assert spec.finding.title == "Oracle attack"


def test_structured_field_with_default() -> None:
    class DefaultStructSpec(PromptSpec):
        """Spec with default structured field."""

        input_documents = ()
        role = SLRole
        task = "Do it"

        finding: Finding = StructuredField(description="Finding", default=Finding(title="default", severity="low"))

    spec = DefaultStructSpec()
    assert spec.finding.title == "default"


def test_structured_field_detection() -> None:
    class DetectSpec(PromptSpec):
        """Doc."""

        input_documents = ()
        role = SLRole
        task = "Do it"

        finding: Finding = StructuredField(description="Finding")
        name: str = Field(description="Name")

    assert _is_structured_field(DetectSpec.model_fields["finding"]) is True
    assert _is_structured_field(DetectSpec.model_fields["name"]) is False


# ---------------------------------------------------------------------------
# ListField: import and declaration
# ---------------------------------------------------------------------------


def test_list_field_importable_from_prompt_compiler() -> None:
    from ai_pipeline_core.prompt_compiler import ListField

    assert callable(ListField)


def test_list_field_importable_from_top_level() -> None:
    from ai_pipeline_core import ListField

    assert callable(ListField)


def test_list_field_creates_valid_spec_with_base_models() -> None:
    class FindingsSpec(PromptSpec):
        """Analyze findings."""

        input_documents = ()
        role = SLRole
        task = "Review the findings."

        findings: list[Finding] = ListField(description="All findings")

    spec = FindingsSpec(findings=[Finding(title="A", severity="high"), Finding(title="B", severity="low")])
    assert len(spec.findings) == 2


def test_list_field_creates_valid_spec_with_strings() -> None:
    class QueriesSpec(PromptSpec):
        """Evaluate queries."""

        input_documents = ()
        role = SLRole
        task = "Evaluate the queries."

        queries: list[str] = ListField(description="Search queries")

    spec = QueriesSpec(queries=["query one", "query two"])
    assert len(spec.queries) == 2


def test_list_field_detection() -> None:
    class DetectSpec(PromptSpec):
        """Doc."""

        input_documents = ()
        role = SLRole
        task = "Do it"

        findings: list[Finding] = ListField(description="Findings")
        name: str = Field(description="Name")

    assert _is_list_field(DetectSpec.model_fields["findings"]) is True
    assert _is_list_field(DetectSpec.model_fields["name"]) is False


# ---------------------------------------------------------------------------
# StructuredField: rendering
# ---------------------------------------------------------------------------


def test_structured_field_renders_json_in_xml() -> None:
    class RenderStructSpec(PromptSpec):
        """Doc."""

        input_documents = ()
        role = SLRole
        task = "Analyze the finding."

        finding: Finding = StructuredField(description="Primary finding")

    spec = RenderStructSpec(finding=Finding(title="Oracle attack", severity="critical"))
    messages = render_multi_line_messages(spec)

    assert len(messages) == 1
    field_name, xml_block = messages[0]
    assert field_name == "finding"
    assert "<finding>" in xml_block
    assert "</finding>" in xml_block
    assert '"title": "Oracle attack"' in xml_block
    assert '"severity": "critical"' in xml_block


def test_structured_field_context_placeholder() -> None:
    class CtxStructSpec(PromptSpec):
        """Doc."""

        input_documents = ()
        role = SLRole
        task = "Do it"

        finding: Finding = StructuredField(description="Primary finding")

    rendered = render_text(CtxStructSpec(finding=Finding(title="A", severity="high")))
    assert "**Primary finding:** (provided in <finding> tags in previous message)" in rendered
    # The JSON should NOT be inline
    assert '"title": "A"' not in rendered


def test_structured_field_nested_model() -> None:
    class NestedSpec(PromptSpec):
        """Doc."""

        input_documents = ()
        role = SLRole
        task = "Do it"

        detail: NestedDetail = StructuredField(description="Detail")

    spec = NestedSpec(detail=NestedDetail(category="infra", tags=["aws", "k8s"]))
    messages = render_multi_line_messages(spec)
    assert len(messages) == 1
    _, xml_block = messages[0]
    assert '"category": "infra"' in xml_block
    assert '"tags"' in xml_block
    assert '"aws"' in xml_block


# ---------------------------------------------------------------------------
# ListField: rendering (BaseModel items)
# ---------------------------------------------------------------------------


def test_list_field_renders_items_in_xml() -> None:
    class RenderListSpec(PromptSpec):
        """Doc."""

        input_documents = ()
        role = SLRole
        task = "Review findings."

        findings: list[Finding] = ListField(description="All findings")

    spec = RenderListSpec(
        findings=[
            Finding(title="A", severity="high"),
            Finding(title="B", severity="low"),
            Finding(title="C", severity="medium"),
        ]
    )
    messages = render_multi_line_messages(spec)

    assert len(messages) == 1
    field_name, xml_block = messages[0]
    assert field_name == "findings"
    assert "<findings>" in xml_block
    assert "</findings>" in xml_block
    assert "<item_1>" in xml_block
    assert "</item_1>" in xml_block
    assert "<item_2>" in xml_block
    assert "<item_3>" in xml_block
    assert '"title": "A"' in xml_block
    assert '"title": "B"' in xml_block
    assert '"title": "C"' in xml_block


def test_list_field_context_placeholder_shows_count() -> None:
    class CountSpec(PromptSpec):
        """Doc."""

        input_documents = ()
        role = SLRole
        task = "Do it"

        findings: list[Finding] = ListField(description="All findings")

    rendered = render_text(
        CountSpec(
            findings=[
                Finding(title="A", severity="high"),
                Finding(title="B", severity="low"),
            ]
        )
    )
    assert "**All findings:** (2 items provided in <findings> tags in previous message)" in rendered


def test_list_field_single_item() -> None:
    class SingleSpec(PromptSpec):
        """Doc."""

        input_documents = ()
        role = SLRole
        task = "Do it"

        findings: list[Finding] = ListField(description="Findings")

    spec = SingleSpec(findings=[Finding(title="Only", severity="low")])
    messages = render_multi_line_messages(spec)
    _, xml_block = messages[0]
    assert "<item_1>" in xml_block
    assert "<item_2>" not in xml_block

    rendered = render_text(spec)
    assert "(1 items provided in <findings> tags in previous message)" in rendered


# ---------------------------------------------------------------------------
# ListField: rendering (string items)
# ---------------------------------------------------------------------------


def test_list_field_string_items() -> None:
    class StringListSpec(PromptSpec):
        """Doc."""

        input_documents = ()
        role = SLRole
        task = "Evaluate queries."

        queries: list[str] = ListField(description="Search queries")

    spec = StringListSpec(queries=["defi oracle", "flash loan exploit"])
    messages = render_multi_line_messages(spec)

    assert len(messages) == 1
    _, xml_block = messages[0]
    assert "<queries>" in xml_block
    assert "<item_1>\ndefi oracle\n</item_1>" in xml_block
    assert "<item_2>\nflash loan exploit\n</item_2>" in xml_block


# ---------------------------------------------------------------------------
# ListField: empty list
# ---------------------------------------------------------------------------


def test_list_field_empty_list() -> None:
    class EmptyListSpec(PromptSpec):
        """Doc."""

        input_documents = ()
        role = SLRole
        task = "Do it"

        items: list[str] = ListField(description="Items", default=[])

    spec = EmptyListSpec()
    messages = render_multi_line_messages(spec)

    assert len(messages) == 1
    _, xml_block = messages[0]
    assert "<items>" in xml_block
    assert "</items>" in xml_block
    assert "<item_1>" not in xml_block

    rendered = render_text(spec)
    assert "(0 items provided in <items> tags in previous message)" in rendered


# ---------------------------------------------------------------------------
# Mixed fields: coexistence
# ---------------------------------------------------------------------------


@pytest.mark.ai_docs
def test_mixed_field_types_coexist() -> None:
    """All four field types on one PromptSpec: Field, MultiLineField, StructuredField, ListField."""
    from ai_pipeline_core.prompt_compiler import MultiLineField

    class MixedSpec(PromptSpec):
        """Mixed field types."""

        input_documents = ()
        role = SLRole
        task = "Process all inputs."

        name: str = Field(description="Project name")
        feedback: str = MultiLineField(description="Review feedback")
        finding: Finding = StructuredField(description="Primary finding")
        findings: list[Finding] = ListField(description="All findings")

    spec = MixedSpec(
        name="ACME",
        feedback="Great work\non this",
        finding=Finding(title="A", severity="high"),
        findings=[Finding(title="B", severity="low")],
    )

    rendered = render_text(spec)
    # Regular field inlined
    assert "**Project name:**\nACME" in rendered
    # MultiLineField placeholder
    assert "(provided in <feedback> tags in previous message)" in rendered
    # StructuredField placeholder
    assert "(provided in <finding> tags in previous message)" in rendered
    # ListField placeholder with count
    assert "(1 items provided in <findings> tags in previous message)" in rendered

    messages = render_multi_line_messages(spec)
    field_names = [name for name, _ in messages]
    assert field_names == ["feedback", "finding", "findings"]


def test_mixed_fields_preserve_declaration_order() -> None:
    class OrderSpec(PromptSpec):
        """Doc."""

        input_documents = ()
        role = SLRole
        task = "Do it"

        alpha: Finding = StructuredField(description="Alpha")
        beta: list[str] = ListField(description="Beta")
        gamma: str = Field(description="Gamma")

    spec = OrderSpec(
        alpha=Finding(title="A", severity="high"),
        beta=["b1", "b2"],
        gamma="short",
    )
    messages = render_multi_line_messages(spec)
    assert [name for name, _ in messages] == ["alpha", "beta"]


# ---------------------------------------------------------------------------
# Preview rendering
# ---------------------------------------------------------------------------


def test_preview_structured_field() -> None:
    class PreviewStructSpec(PromptSpec):
        """Doc."""

        input_documents = ()
        role = SLRole
        task = "Analyze it"

        finding: Finding = StructuredField(description="Finding")

    preview = render_preview(PreviewStructSpec)
    assert "<finding>{finding}</finding>" in preview
    assert "---" in preview


def test_preview_list_field() -> None:
    class PreviewListSpec(PromptSpec):
        """Doc."""

        input_documents = ()
        role = SLRole
        task = "Review items"

        findings: list[Finding] = ListField(description="Findings")

    preview = render_preview(PreviewListSpec)
    assert "<findings>{findings}</findings>" in preview
    assert "---" in preview


def test_preview_no_separator_without_special_fields() -> None:
    class PlainSpec(PromptSpec):
        """Doc."""

        input_documents = ()
        role = SLRole
        task = "Do it"

        name: str = Field(description="Name")

    preview = render_preview(PlainSpec)
    assert "---" not in preview


# ---------------------------------------------------------------------------
# Conversation.send_spec integration
# ---------------------------------------------------------------------------


@pytest.mark.asyncio
async def test_send_spec_structured_field_as_user_message() -> None:
    class SendStructSpec(PromptSpec):
        """Doc."""

        input_documents = ()
        role = SLRole
        task = "Analyze the finding."

        finding: Finding = StructuredField(description="Finding")

    response = _make_model_response("analysis done")
    conv_after = Conversation[str](model="gpt-5", messages=(response,))

    sent_content: list[str] = []

    async def fake_send(self, content, *, purpose=None, expected_cost=None, **kwargs):
        if isinstance(content, str):
            sent_content.append(content)
        return conv_after

    with patch.object(Conversation, "send", fake_send):
        conv = Conversation[str](model="gpt-5")
        spec = SendStructSpec(finding=Finding(title="Test", severity="high"))
        await conv.send_spec(spec)

    assert len(sent_content) >= 1
    prompt_text = sent_content[-1]
    # JSON should not be in the prompt text
    assert '"title": "Test"' not in prompt_text
    assert "provided in <finding> tags in previous message" in prompt_text


@pytest.mark.asyncio
async def test_send_spec_list_field_as_user_message() -> None:
    class SendListSpec(PromptSpec):
        """Doc."""

        input_documents = ()
        role = SLRole
        task = "Review findings."

        findings: list[Finding] = ListField(description="Findings")

    response = _make_model_response("review done")
    messages_at_send: list[tuple[object, ...]] = []

    async def fake_send(self, content, *, purpose=None, expected_cost=None, **kwargs):
        messages_at_send.append(self.messages)
        return Conversation[str](model="gpt-5", messages=(*self.messages, response))

    with patch.object(Conversation, "send", fake_send):
        conv = Conversation[str](model="gpt-5")
        spec = SendListSpec(findings=[Finding(title="A", severity="high"), Finding(title="B", severity="low")])
        await conv.send_spec(spec)

    assert len(messages_at_send) == 1
    msgs = messages_at_send[0]
    msg_texts = [m.text for m in msgs if hasattr(m, "text")]
    xml_msgs = [t for t in msg_texts if "<findings>" in t]
    assert len(xml_msgs) == 1
    assert "<item_1>" in xml_msgs[0]
    assert "<item_2>" in xml_msgs[0]


@pytest.mark.asyncio
async def test_send_spec_all_field_types_combined() -> None:
    from ai_pipeline_core.prompt_compiler import MultiLineField

    class CombinedSpec(PromptSpec):
        """Doc."""

        input_documents = ()
        role = SLRole
        task = "Process everything."

        name: str = Field(description="Name")
        notes: str = MultiLineField(description="Notes")
        finding: Finding = StructuredField(description="Finding")
        findings: list[Finding] = ListField(description="Findings")

    response = _make_model_response("done")
    messages_at_send: list[tuple[object, ...]] = []

    async def fake_send(self, content, *, purpose=None, expected_cost=None, **kwargs):
        messages_at_send.append(self.messages)
        return Conversation[str](model="gpt-5", messages=(*self.messages, response))

    with patch.object(Conversation, "send", fake_send):
        conv = Conversation[str](model="gpt-5")
        spec = CombinedSpec(
            name="ACME",
            notes="some notes",
            finding=Finding(title="X", severity="low"),
            findings=[Finding(title="Y", severity="high")],
        )
        await conv.send_spec(spec)

    assert len(messages_at_send) == 1
    msgs = messages_at_send[0]
    msg_texts = [m.text for m in msgs if hasattr(m, "text")]
    # All three special fields combined in a single user message
    combined = [t for t in msg_texts if "<notes>" in t or "<finding>" in t or "<findings>" in t]
    assert len(combined) == 1
    assert "<notes>some notes</notes>" in combined[0]
    assert "<finding>" in combined[0]
    assert "<findings>" in combined[0]
    assert "<item_1>" in combined[0]


# ---------------------------------------------------------------------------
# Edge cases
# ---------------------------------------------------------------------------


def test_structured_field_requires_description() -> None:
    with pytest.raises(TypeError, match=r"must use Field\(description='\.\.\.'\)|description"):

        class BadSpec(PromptSpec):
            """Doc."""

            input_documents = ()
            role = SLRole
            task = "Do it"

            finding: Finding = StructuredField()  # type: ignore[call-arg]


def test_list_field_requires_description() -> None:
    with pytest.raises(TypeError, match=r"must use Field\(description='\.\.\.'\)|description"):

        class BadSpec(PromptSpec):
            """Doc."""

            input_documents = ()
            role = SLRole
            task = "Do it"

            items: list[str] = ListField()  # type: ignore[call-arg]


def test_structured_field_frozen_spec() -> None:
    class FrozenStructSpec(PromptSpec):
        """Doc."""

        input_documents = ()
        role = SLRole
        task = "Do it"

        finding: Finding = StructuredField(description="Finding")

    spec = FrozenStructSpec(finding=Finding(title="A", severity="high"))
    with pytest.raises(Exception):
        spec.finding = Finding(title="B", severity="low")  # type: ignore[misc]


def test_list_field_with_nested_models() -> None:
    class NestedListSpec(PromptSpec):
        """Doc."""

        input_documents = ()
        role = SLRole
        task = "Do it"

        details: list[NestedDetail] = ListField(description="Details")

    spec = NestedListSpec(
        details=[
            NestedDetail(category="infra", tags=["aws", "k8s"]),
            NestedDetail(category="app", tags=["python"]),
        ]
    )
    messages = render_multi_line_messages(spec)
    _, xml_block = messages[0]
    assert "<item_1>" in xml_block
    assert '"category": "infra"' in xml_block
    assert '"tags"' in xml_block
    assert "<item_2>" in xml_block
    assert '"category": "app"' in xml_block


def test_follow_up_spec_with_structured_field() -> None:
    class InitSpec(PromptSpec):
        """Init."""

        input_documents = ()
        role = SLRole
        task = "Start"

    class FollowStructSpec(PromptSpec, follows=InitSpec):
        """Follow with structured."""

        task = "Continue with finding."

        finding: Finding = StructuredField(description="Finding")

    spec = FollowStructSpec(finding=Finding(title="A", severity="high"))
    rendered = render_text(spec)
    assert "# Role" not in rendered
    assert "(provided in <finding> tags in previous message)" in rendered

    messages = render_multi_line_messages(spec)
    assert len(messages) == 1
    assert messages[0][0] == "finding"


def test_follow_up_spec_with_list_field() -> None:
    class InitSpec(PromptSpec):
        """Init."""

        input_documents = ()
        role = SLRole
        task = "Start"

    class FollowListSpec(PromptSpec, follows=InitSpec):
        """Follow with list."""

        task = "Continue with findings."

        findings: list[Finding] = ListField(description="Findings")

    spec = FollowListSpec(findings=[Finding(title="A", severity="high")])
    rendered = render_text(spec)
    assert "(1 items provided in <findings> tags in previous message)" in rendered


def test_structured_field_json_is_indented() -> None:
    """Verify the JSON output uses indent=2 for readability."""

    class IndentSpec(PromptSpec):
        """Doc."""

        input_documents = ()
        role = SLRole
        task = "Do it"

        finding: Finding = StructuredField(description="Finding")

    spec = IndentSpec(finding=Finding(title="Test", severity="high"))
    messages = render_multi_line_messages(spec)
    _, xml_block = messages[0]
    # indent=2 means fields should be on separate lines with 2-space indent
    assert '  "title": "Test"' in xml_block
    assert '  "severity": "high"' in xml_block


def test_list_field_basemodel_items_json_is_indented() -> None:
    """Verify BaseModel items in ListField use indent=2 JSON."""

    class IndentListSpec(PromptSpec):
        """Doc."""

        input_documents = ()
        role = SLRole
        task = "Do it"

        findings: list[Finding] = ListField(description="Findings")

    spec = IndentListSpec(findings=[Finding(title="Test", severity="high")])
    messages = render_multi_line_messages(spec)
    _, xml_block = messages[0]
    assert '  "title": "Test"' in xml_block



"""Reconstruct lifecycle events from database span trees."""

import json
from dataclasses import dataclass
from datetime import datetime
from typing import Any
from uuid import UUID

from ai_pipeline_core._lifecycle_events import DocumentRef, TaskCompletedEvent, TaskFailedEvent, TaskStartedEvent
from ai_pipeline_core.database._json_helpers import parse_json_object
from ai_pipeline_core.database._protocol import DatabaseReader
from ai_pipeline_core.database._types import DocumentRecord, SpanKind, SpanRecord, SpanStatus

from ._event_serialization import event_to_payload
from ._types import (
    ErrorCode,
    EventType,
    FlowCompletedEvent,
    FlowFailedEvent,
    FlowSkippedEvent,
    FlowStartedEvent,
    RunCompletedEvent,
    RunFailedEvent,
    RunStartedEvent,
)

__all__ = [
    "_ReconstructedEvent",
    "_reconstruct_lifecycle_events",
]

_LIFECYCLE_KINDS = {SpanKind.DEPLOYMENT, SpanKind.FLOW, SpanKind.TASK}

_KIND_PRIORITY = {SpanKind.DEPLOYMENT: 0, SpanKind.FLOW: 1, SpanKind.TASK: 2}
_PHASE_PRIORITY_STARTED = 0
_PHASE_PRIORITY_SKIPPED = 1
_PHASE_PRIORITY_TERMINAL = 2


@dataclass(frozen=True, slots=True)
class _ReconstructedEvent:
    """A lifecycle event reconstructed from database spans."""

    event_type: EventType
    span_id: str
    timestamp: datetime
    data: dict[str, Any]


def _parse_meta(span: SpanRecord) -> dict[str, Any]:
    return parse_json_object(span.meta_json, context=f"Span {span.span_id}", field_name="meta_json")


def _parse_receiver(span: SpanRecord) -> dict[str, Any]:
    if not span.receiver_json:
        return {}
    parsed = parse_json_object(span.receiver_json, context=f"Span {span.span_id}", field_name="receiver_json")
    return parsed.get("value", {}) if isinstance(parsed.get("value"), dict) else {}


def _parse_output(span: SpanRecord) -> dict[str, Any]:
    if not span.output_json:
        return {}
    try:
        value = json.loads(span.output_json)
    except json.JSONDecodeError:
        return {}
    return value if isinstance(value, dict) else {}


def _duration_ms(span: SpanRecord) -> int:
    if span.started_at and span.ended_at:
        return int((span.ended_at - span.started_at).total_seconds() * 1000)
    return 0


def _class_name_from_target(target: str) -> str:
    if not target:
        return ""
    if ":" in target:
        qual_part = target.rsplit(":", 1)[-1]
        parts = qual_part.split(".")
        return parts[0] if len(parts) >= 2 else qual_part
    return target.rsplit(".", 1)[-1]


def _build_doc_refs(shas: tuple[str, ...], doc_map: dict[str, DocumentRecord]) -> tuple[DocumentRef, ...]:
    refs: list[DocumentRef] = []
    for sha in shas:
        record = doc_map.get(sha)
        if record is not None:
            refs.append(DocumentRef.from_record(record))
    return tuple(refs)


def _sort_key(event: _ReconstructedEvent) -> tuple[datetime, int, int, str]:
    kind_priority = 1
    phase_priority = _PHASE_PRIORITY_STARTED
    event_type = event.event_type

    if event_type in {EventType.RUN_STARTED, EventType.RUN_COMPLETED, EventType.RUN_FAILED}:
        kind_priority = _KIND_PRIORITY[SpanKind.DEPLOYMENT]
    elif event_type in {EventType.FLOW_STARTED, EventType.FLOW_COMPLETED, EventType.FLOW_FAILED, EventType.FLOW_SKIPPED}:
        kind_priority = _KIND_PRIORITY[SpanKind.FLOW]
    elif event_type in {EventType.TASK_STARTED, EventType.TASK_COMPLETED, EventType.TASK_FAILED}:
        kind_priority = _KIND_PRIORITY[SpanKind.TASK]

    if event_type in {EventType.FLOW_SKIPPED}:
        phase_priority = _PHASE_PRIORITY_SKIPPED
    elif event_type in {
        EventType.RUN_COMPLETED,
        EventType.RUN_FAILED,
        EventType.FLOW_COMPLETED,
        EventType.FLOW_FAILED,
        EventType.TASK_COMPLETED,
        EventType.TASK_FAILED,
    }:
        phase_priority = _PHASE_PRIORITY_TERMINAL

    return (event.timestamp, kind_priority, phase_priority, event.span_id)


def _reconstruct_deployment_events(
    span: SpanRecord,
    meta: dict[str, Any],
    parent_task_id_str: str | None,
) -> list[_ReconstructedEvent]:
    events: list[_ReconstructedEvent] = []
    span_id_str = str(span.span_id)
    root_id_str = str(span.root_deployment_id)

    started_event = RunStartedEvent(
        run_id=span.run_id,
        span_id=span_id_str,
        root_deployment_id=root_id_str,
        parent_deployment_task_id=parent_task_id_str,
        input_fingerprint=meta.get("input_fingerprint", ""),
        status=str(SpanStatus.RUNNING),
        deployment_name=span.deployment_name or span.name,
        deployment_class=meta.get("deployment_class", ""),
        flow_plan=meta.get("flow_plan", []),
        parent_span_id=str(span.parent_span_id) if span.parent_span_id else "",
        input_document_sha256s=span.input_document_shas,
    )
    events.append(
        _ReconstructedEvent(
            event_type=EventType.RUN_STARTED,
            span_id=span_id_str,
            timestamp=span.started_at,
            data=event_to_payload(started_event),
        )
    )

    if span.status == SpanStatus.COMPLETED:
        output = _parse_output(span)
        events.append(
            _ReconstructedEvent(
                event_type=EventType.RUN_COMPLETED,
                span_id=span_id_str,
                timestamp=span.ended_at or span.started_at,
                data=event_to_payload(
                    RunCompletedEvent(
                        run_id=span.run_id,
                        span_id=span_id_str,
                        root_deployment_id=root_id_str,
                        parent_deployment_task_id=parent_task_id_str,
                        status=str(span.status),
                        result=output.get("result", {}),
                        deployment_name=span.deployment_name or span.name,
                        deployment_class=meta.get("deployment_class", ""),
                        duration_ms=_duration_ms(span),
                        output_document_sha256s=span.output_document_shas,
                        parent_span_id=str(span.parent_span_id) if span.parent_span_id else "",
                    )
                ),
            )
        )
    elif span.status == SpanStatus.FAILED:
        error_code_str = meta.get("error_code", "unknown")
        try:
            error_code = ErrorCode(error_code_str)
        except ValueError:
            error_code = ErrorCode.UNKNOWN
        events.append(
            _ReconstructedEvent(
                event_type=EventType.RUN_FAILED,
                span_id=span_id_str,
                timestamp=span.ended_at or span.started_at,
                data=event_to_payload(
                    RunFailedEvent(
                        run_id=span.run_id,
                        span_id=span_id_str,
                        root_deployment_id=root_id_str,
                        parent_deployment_task_id=parent_task_id_str,
                        status=str(span.status),
                        error_code=error_code,
                        error_message=span.error_message,
                        deployment_name=span.deployment_name or span.name,
                        deployment_class=meta.get("deployment_class", ""),
                        duration_ms=_duration_ms(span),
                        parent_span_id=str(span.parent_span_id) if span.parent_span_id else "",
                    )
                ),
            )
        )

    return events


def _reconstruct_flow_events(
    span: SpanRecord,
    meta: dict[str, Any],
    parent_task_id_str: str | None,
    deployment_span_id: str,
    doc_map: dict[str, DocumentRecord],
) -> list[_ReconstructedEvent]:
    events: list[_ReconstructedEvent] = []
    span_id_str = str(span.span_id)
    root_id_str = str(span.root_deployment_id)
    flow_class = meta.get("flow_class") or _class_name_from_target(span.target)
    step = meta.get("step", 0)
    total_steps = meta.get("total_steps", 0)

    if span.status in {SpanStatus.SKIPPED, SpanStatus.CACHED}:
        events.append(
            _ReconstructedEvent(
                event_type=EventType.FLOW_SKIPPED,
                span_id=span_id_str,
                timestamp=span.started_at,
                data=event_to_payload(
                    FlowSkippedEvent(
                        run_id=span.run_id,
                        span_id=span_id_str,
                        root_deployment_id=root_id_str,
                        parent_deployment_task_id=parent_task_id_str,
                        flow_name=span.name,
                        flow_class=flow_class,
                        step=step,
                        total_steps=total_steps,
                        status=str(span.status),
                        reason=meta.get("skip_reason", ""),
                        parent_span_id=deployment_span_id,
                        input_document_sha256s=span.input_document_shas,
                    )
                ),
            )
        )
        return events

    flow_params = _parse_receiver(span)
    started_event = FlowStartedEvent(
        run_id=span.run_id,
        span_id=span_id_str,
        root_deployment_id=root_id_str,
        parent_deployment_task_id=parent_task_id_str,
        flow_name=span.name,
        flow_class=flow_class,
        step=step,
        total_steps=total_steps,
        status=str(SpanStatus.RUNNING),
        expected_tasks=meta.get("expected_task_names", []),
        flow_params=flow_params,
        parent_span_id=deployment_span_id,
        input_document_sha256s=span.input_document_shas,
    )
    events.append(
        _ReconstructedEvent(
            event_type=EventType.FLOW_STARTED,
            span_id=span_id_str,
            timestamp=span.started_at,
            data=event_to_payload(started_event),
        )
    )

    if span.status == SpanStatus.COMPLETED:
        events.append(
            _ReconstructedEvent(
                event_type=EventType.FLOW_COMPLETED,
                span_id=span_id_str,
                timestamp=span.ended_at or span.started_at,
                data=event_to_payload(
                    FlowCompletedEvent(
                        run_id=span.run_id,
                        span_id=span_id_str,
                        root_deployment_id=root_id_str,
                        parent_deployment_task_id=parent_task_id_str,
                        flow_name=span.name,
                        flow_class=flow_class,
                        step=step,
                        total_steps=total_steps,
                        status=str(span.status),
                        duration_ms=_duration_ms(span),
                        output_documents=_build_doc_refs(span.output_document_shas, doc_map),
                        parent_span_id=deployment_span_id,
                        input_document_sha256s=span.input_document_shas,
                    )
                ),
            )
        )
    elif span.status == SpanStatus.FAILED:
        events.append(
            _ReconstructedEvent(
                event_type=EventType.FLOW_FAILED,
                span_id=span_id_str,
                timestamp=span.ended_at or span.started_at,
                data=event_to_payload(
                    FlowFailedEvent(
                        run_id=span.run_id,
                        span_id=span_id_str,
                        root_deployment_id=root_id_str,
                        parent_deployment_task_id=parent_task_id_str,
                        flow_name=span.name,
                        flow_class=flow_class,
                        step=step,
                        total_steps=total_steps,
                        status=str(span.status),
                        error_message=span.error_message,
                        parent_span_id=deployment_span_id,
                        input_document_sha256s=span.input_document_shas,
                    )
                ),
            )
        )

    return events


def _reconstruct_task_events(
    span: SpanRecord,
    meta: dict[str, Any],
    parent_task_id_str: str | None,
    *,
    flow_span: SpanRecord | None,
    flow_meta: dict[str, Any],
    doc_map: dict[str, DocumentRecord],
) -> list[_ReconstructedEvent]:
    events: list[_ReconstructedEvent] = []
    span_id_str = str(span.span_id)
    root_id_str = str(span.root_deployment_id)
    task_class = meta.get("task_class") or _class_name_from_target(span.target)
    flow_name = flow_span.name if flow_span else ""
    step = flow_meta.get("step", 0) if flow_meta else 0
    total_steps = flow_meta.get("total_steps", 0) if flow_meta else 0
    parent_span_id = str(span.parent_span_id) if span.parent_span_id else ""

    is_cached = span.status == SpanStatus.CACHED

    if not is_cached:
        events.append(
            _ReconstructedEvent(
                event_type=EventType.TASK_STARTED,
                span_id=span_id_str,
                timestamp=span.started_at,
                data=event_to_payload(
                    TaskStartedEvent(
                        run_id=span.run_id,
                        span_id=span_id_str,
                        root_deployment_id=root_id_str,
                        parent_deployment_task_id=parent_task_id_str,
                        flow_name=flow_name,
                        step=step,
                        total_steps=total_steps,
                        status=str(SpanStatus.RUNNING),
                        task_name=span.name,
                        task_class=task_class,
                        parent_span_id=parent_span_id,
                        input_document_sha256s=span.input_document_shas,
                    )
                ),
            )
        )

    if span.status in {SpanStatus.COMPLETED, SpanStatus.CACHED}:
        events.append(
            _ReconstructedEvent(
                event_type=EventType.TASK_COMPLETED,
                span_id=span_id_str,
                timestamp=span.ended_at or span.started_at,
                data=event_to_payload(
                    TaskCompletedEvent(
                        run_id=span.run_id,
                        span_id=span_id_str,
                        root_deployment_id=root_id_str,
                        parent_deployment_task_id=parent_task_id_str,
                        flow_name=flow_name,
                        step=step,
                        total_steps=total_steps,
                        status="cached" if is_cached else str(span.status),
                        task_name=span.name,
                        task_class=task_class,
                        duration_ms=_duration_ms(span),
                        output_documents=_build_doc_refs(span.output_document_shas, doc_map),
                        parent_span_id=parent_span_id,
                        input_document_sha256s=span.input_document_shas,
                    )
                ),
            )
        )
    elif span.status == SpanStatus.FAILED:
        events.append(
            _ReconstructedEvent(
                event_type=EventType.TASK_FAILED,
                span_id=span_id_str,
                timestamp=span.ended_at or span.started_at,
                data=event_to_payload(
                    TaskFailedEvent(
                        run_id=span.run_id,
                        span_id=span_id_str,
                        root_deployment_id=root_id_str,
                        parent_deployment_task_id=parent_task_id_str,
                        flow_name=flow_name,
                        step=step,
                        total_steps=total_steps,
                        status=str(span.status),
                        task_name=span.name,
                        task_class=task_class,
                        error_message=span.error_message,
                        parent_span_id=parent_span_id,
                        input_document_sha256s=span.input_document_shas,
                    )
                ),
            )
        )

    return events


async def _reconstruct_lifecycle_events(
    reader: DatabaseReader,
    root_deployment_id: UUID,
) -> list[_ReconstructedEvent]:
    """Reconstruct lifecycle events from a deployment's span tree.

    Returns events in chronological order. Each event's data dict matches
    the payload shape published by PubSubPublisher (via event_to_payload()).

    Heartbeats are not reconstructed (ephemeral, not stored in spans).
    """
    all_spans = await reader.get_deployment_tree(root_deployment_id)

    lifecycle_spans = [s for s in all_spans if s.kind in _LIFECYCLE_KINDS]
    if not lifecycle_spans:
        return []

    all_doc_shas: set[str] = set()
    for span in lifecycle_spans:
        all_doc_shas.update(span.input_document_shas)
        all_doc_shas.update(span.output_document_shas)

    doc_map: dict[str, DocumentRecord] = {}
    if all_doc_shas:
        doc_map = await reader.get_documents_batch(sorted(all_doc_shas))

    span_by_id: dict[UUID, SpanRecord] = {s.span_id: s for s in lifecycle_spans}
    meta_by_id: dict[UUID, dict[str, Any]] = {s.span_id: _parse_meta(s) for s in lifecycle_spans}

    deployment_span: SpanRecord | None = None
    for span in lifecycle_spans:
        if span.kind == SpanKind.DEPLOYMENT:
            deployment_span = span
            break

    parent_task_id_str: str | None = None
    deployment_span_id_str = ""
    if deployment_span is not None:
        parent_task_id_str = str(deployment_span.parent_span_id) if deployment_span.parent_span_id else None
        deployment_span_id_str = str(deployment_span.span_id)

    events: list[_ReconstructedEvent] = []

    for span in lifecycle_spans:
        meta = meta_by_id[span.span_id]

        if span.kind == SpanKind.DEPLOYMENT:
            events.extend(_reconstruct_deployment_events(span, meta, parent_task_id_str))

        elif span.kind == SpanKind.FLOW:
            events.extend(_reconstruct_flow_events(span, meta, parent_task_id_str, deployment_span_id_str, doc_map))

        elif span.kind == SpanKind.TASK:
            flow_span = span_by_id.get(span.parent_span_id) if span.parent_span_id else None
            flow_meta = meta_by_id.get(span.parent_span_id, {}) if span.parent_span_id else {}
            events.extend(_reconstruct_task_events(span, meta, parent_task_id_str, flow_span=flow_span, flow_meta=flow_meta, doc_map=doc_map))

    events.sort(key=_sort_key)
    return events
